package org.ugasp.connector.manager.impl;


import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import org.ugasp.core.ActorSession;
import org.ugasp.core.ApplicationInstance;
import org.ugasp.core.Platform;
import org.ugasp.logging.Logger;
import org.ugasp.connector.manager.DataTransmissionRequestManager;
import org.ugasp.service.external.player.PlayerConnector;

/**
 * This class manages the treatment of data received from and sent to a client,
 * using the first version of the data transmission protocol (including handling of MooDS and
 * RMAI protocols).
 * 
 * @author Gabriel Adgeg (gabriel.adgeg@telecom-sudparis.eu)
 */
public class DataTransmissionRequestManager_v1 implements DataTransmissionRequestManager{
	
	public static final byte 			DATA_TRANSMISSION_PROTOCOL_VERSION = 1;
	// Client request
	private static final byte 			ADMIN_REQUEST_TYPE = 0;
    private static final byte 			GAME_MASTER_REQUEST_TYPE = 1;
    private static final byte 			PLAYER_REQUEST_TYPE = 2;
    private static final byte 			SENSOR_REQUEST_TYPE = 3;
    // Player requests
	private static final byte 			GET_EVENTS_REQUEST_TYPE = 0;
	private static final byte 			RMAI_REQUEST_TYPE = 1;
	private static final byte 			MOODS_REQUEST_TYPE = 2;
	// This attributes matches with the additional length of the protocol header, use to inform about the full size of message
	private static final int 			HEADER_ADDITIONAL_LENGTH = 3;
	
	private DataInputStream 			requestInputStream;
	private DataOutputStream 			responseOutputStream;
	private Platform					core;
	private PlayerConnector 			playerConnector;
	
	
    public DataTransmissionRequestManager_v1(DataInputStream requestInputStream, DataOutputStream responseOutputStream, Platform core, PlayerConnector playerConnector){
    	this.requestInputStream 		= requestInputStream;
    	this.responseOutputStream 		= responseOutputStream;
    	this.core					 	= core;
    	this.playerConnector			= playerConnector;
    }
	
    /**
	 * Get the DataInputStream matching with the client request.
	 * 
	 * @return 	the DataInputStream
	 */
	public DataInputStream getRequestInputStream(){
		return requestInputStream;
	}
	
	/**
	 * Get the DataOutputStream matching with the platform response.
	 * 
	 * @return 	the DataOutputStream
	 */
	public DataOutputStream getResponseOutputStream(){
		return responseOutputStream;
	}
	
	/**
	 * Get the PlayerConnector needed to:
	 * <ul>
	 * <li>treat data received from a player</li>
	 * <li>transmit data designed to this player</li>
	 * </ul>
	 * @return 	the PlayerConnector
	 */
	public PlayerConnector getPlayerConnector(){
		return playerConnector;
	}
	
	/**
	 * Get the platform. Needed to retrieve an ApplicationInstance from an actorSessionID.
	 * 
	 * @return 	the uGASP Platform
	 */
	public Platform getPlatform(){
		return core;
	}
	
	/**
	 * Treat the data transmission request sent by a client.
	 * Treatment of the request includes the suitable management for the data sent by the client,
	 * and the sending of events designed to the client on the response of its request.
	 * 
	 * @throws IOException	if a problem occurs on the reading of the request
	 * 						or on the writing of the response
	 */
	public void treatClientRequest() throws IOException{
		core.log(Logger.TRACE,this.getClass(),"treatClientRequest method called");
		// retrieve the requester type
    	byte requestType = requestInputStream.readByte();
    	switch(requestType){
    		case ADMIN_REQUEST_TYPE:
    			treatAdminRequest();
    			break;
    			
    		case GAME_MASTER_REQUEST_TYPE:
    			treatGameMasterRequest();
    			break;
    			
     		case PLAYER_REQUEST_TYPE:
     			short senderActorSessionID = requestInputStream.readShort();
     			treatPlayerRequest(senderActorSessionID);
     			break;
     			
     		case SENSOR_REQUEST_TYPE:
     			treatSensorRequest();
     			break;
     			
     		default:
     			core.log(Logger.DEBUG, this.getClass(),"Uncorrect request type received");
     			break;
    	}
	}
	
	/*
	 * Treat data included in a player request. On the first version of the data transmission protocol,
	 * data could be either MooDS object or RMAI.
	 * Just after the treatment of the player's data, this method also send events designed to player, if so.
	 *  
	 */
	private void treatPlayerRequest(short playerActorSessionID) throws IOException{
		core.log(Logger.TRACE,this.getClass(),"treatPlayerRequest method called for actor session with ID = "+playerActorSessionID);
		// retrieve actorSession from actorSessionID
		ActorSession actorSession = getActorSessionFromActorSessionID(playerActorSessionID);
		if(actorSession != null){
			// check ID of player message and ID last message acknowledged   
			checkMessagesIDsFromPlayerRequest(actorSession);
			byte requesterType = requestInputStream.readByte();
			while(requesterType != GET_EVENTS_REQUEST_TYPE){
				switch(requesterType){
					case RMAI_REQUEST_TYPE:
						// retrieve recipient and invoked method
						short recipientActorSessionID = requestInputStream.readShort();
						String invokedMethodXMLSerialized = requestInputStream.readUTF();
		     			core.log(Logger.TRACE, this.getClass(),"RMAI request type received by ActorSession with ID = "+playerActorSessionID+" designed for ActorSession with ID = "+recipientActorSessionID);
						playerConnector.remoteMethodAsynchronousInvocation(playerActorSessionID, recipientActorSessionID, invokedMethodXMLSerialized);
						break;
						
					case MOODS_REQUEST_TYPE:
		     			core.log(Logger.TRACE, this.getClass(),"MooDS request type received by ActorSession with ID = "+playerActorSessionID);
						playerConnector.sendMooDSObject(requestInputStream, playerActorSessionID);
						break;
						
					default:
						break;	
				}
				// retrieve requester type
				requesterType = requestInputStream.readByte();
			}
			// send events to the player
			sendEventsToPlayer(actorSession);
		}else{
			core.log(Logger.TRACE,this.getClass(),"treatPlayerRequest method called --> actor session with ID = "+playerActorSessionID+" cannot be retrieved.");
		}
	}

	/*
	 * Put events designed to a player on its send buffer (DataOutputStream), then
	 * write this send buffer (ByteArrayOutputStream) on the response output stream.
	 */
	private void sendEventsToPlayer(ActorSession actorSession) throws IOException{
		try{
			DataOutputStream sendBufferDataOutputStream = actorSession.getSendBufferDataOutputStream();
			if(sendBufferDataOutputStream != null){
				// send events to client designated by its actor session ID within sendBufferDataOutputStream
				playerConnector.getEvents(sendBufferDataOutputStream, actorSession.getActorSessionID());
				// send response to the client
				writeDataOnResponseOutputStream(actorSession);
			}
		}finally{
			if(responseOutputStream != null){
				responseOutputStream.close();
			}
		}

	}

	
	/*
	 * Read IDs of message sent by player.
	 * ID of message sent by player is inserted in lastMessageReceivedIDByActorSessionID hashtable.
	 * ID of message acknowledgment sent by player is compared with ID of last message sent to player.
	 * If they are equals, events designed to the player can be deleted, ID of last message sent to player can be 
	 * update then store into messageSentIDByActorSessionID hashtable.
	 */
	private void checkMessagesIDsFromPlayerRequest(ActorSession actorSession) throws IOException{
		// ID of message received from the client
		byte receivedMessageID = requestInputStream.readByte();
		//dataTransmissionConnector.putLastMesageReceivedID(playerActorSessionID, receivedMessageID);
		actorSession.setLastMesageReceivedID(receivedMessageID);
		// ID of message acknowledged by the client
		byte receivedAckMessageID = requestInputStream.readByte();
		// first message received from client use this typical value to tell it hasn't acknowledged anything yet
		if(receivedAckMessageID != Byte.MIN_VALUE){
			// test whether client has acknowledged message
			if(receivedAckMessageID == actorSession.getMessageSentID()){
				// events can be deleted since client properly received them
				actorSession.getSendBufferByteArrayOutputStream().reset();
				// update messageSentID
				byte messageSentID = (byte)((receivedAckMessageID + 1)%Byte.MAX_VALUE);
				actorSession.setMessageSentID(messageSentID);
			}
		}
	}
	
	
	/*
	 * This method successively:
	 * - write suitable header on the response output stream
	 * - write data to transmit
	 */
	private void writeDataOnResponseOutputStream(ActorSession actorSession) throws IOException{
		// write header of response for this actorSessionID
		writeRequestHeaderOnResponseOutputStream(actorSession);
		// write events on response output stream
		actorSession.getSendBufferByteArrayOutputStream().writeTo(responseOutputStream);
	}

	
	/*
	 * This method writes the header matching with data that are going to be sent to the client on the response output stream.
	 * The header includes:
	 * - the length of the message
	 * - the version of the protocol used by this DataTrasnmissionRequestManager
	 * - the ID of the message which is going to be sent
	 * - the ID of the last message received from the client.
	 */
	private void writeRequestHeaderOnResponseOutputStream(ActorSession actorSession) throws IOException{	
		//write message length
		int messageLength = HEADER_ADDITIONAL_LENGTH + actorSession.getSendBufferByteArrayOutputStream().size();
		responseOutputStream.writeInt(messageLength);
		// protocol version
		responseOutputStream.writeByte(DATA_TRANSMISSION_PROTOCOL_VERSION);
		// write ID of message
		byte messageSentID = actorSession.getMessageSentID();
		responseOutputStream.writeByte(messageSentID);
		// write ID of message last message received from this actorSessionID
		byte lastMessageReceivedID = actorSession.getLastMessageReceivedID();
		core.log(Logger.TRACE,this.getClass(),"Message received from client with actor session ID = "+actorSession.getActorSessionID()+" --> "+lastMessageReceivedID);
		responseOutputStream.writeByte(lastMessageReceivedID);
	}
	
	
	private ActorSession getActorSessionFromActorSessionID(short actorSessionID){
		ApplicationInstance instance = core.getApplicationInstanceFromActorSessionID(actorSessionID);
		if(instance != null){
			return instance.getActorSession(actorSessionID);
		}else{
			return null;
		}
	}
	
	/********** Unimplemented methods ****************/
	
	private void treatAdminRequest(){}
	
	
	private void treatGameMasterRequest(){}
	
	
	private void treatSensorRequest(){}

}
