/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.impl.connector.http;


import org.osgi.framework.BundleContext;
import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Dictionary;
import java.util.Hashtable;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.ugasp.core.Platform;
import org.ugasp.logging.Logger;
import org.ugasp.service.external.admin.AdminConnector;
import org.ugasp.service.external.player.PlayerConnector;


/**
 * This servlet is call the first time the user log in with the Application.
 * The HTTP parameters are :
 * - uid the UserID of the operator
 * - appID the ApplicationID hard coded in the application.
 * - username the choosed username
 * - password the choosed password
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project 
 */
public class ApplicationManagementServlet extends HttpServlet implements org.ugasp.connector.http.Servlet{
	
    //iPojo dynamic binding fields
	private Platform 		core;
	private PlayerConnector playerConnector;
    private AdminConnector 	adminConnector;
    private HttpService 	httpService;

    //other fields
    private DataTransmissionServlet dataTransmissionServlet;
        
    private BundleContext bundleContext;
    
    private static final String SERVLET_URL = "/uGASP/Servlet";
    
    public ApplicationManagementServlet(BundleContext bc){
    	bundleContext = bc;
	    // instantiate BinaryServlet
    	dataTransmissionServlet = new DataTransmissionServlet();
		core.log(Logger.INFO, this.getClass(), "Application Management Servlet instantiated");	
    }
    
    
    public void start(){
    	dataTransmissionServlet.start();
		core.log(Logger.DEBUG, this.getClass(), "Application Management Servlet started");
    }
    
    
    public void stop(){
    	dataTransmissionServlet.stop();
		core.log(Logger.DEBUG, this.getClass(), "Application Management Servlet stopped");
    }
    
    
    public void bindPlatform(Platform core){
    	dataTransmissionServlet.bindPlatform(core);
		core.log(Logger.DEBUG, this.getClass(), "Application Management bind with ugasp platform");
    }
    
    
    public void unbindPlatform(){
    	if(core != null){
	    	core.log(Logger.DEBUG, this.getClass(), "Application Management Servlet unbind with ugasp platform");
	    	dataTransmissionServlet.unbindPlatform();
    	}
    }
    
    
    public void bindPlayerConnector(PlayerConnector connector){
    	dataTransmissionServlet.bindPlayerConnector(connector);
		core.log(Logger.DEBUG, this.getClass(), "Application Management Servlet bind with PlayerConnector");
    }
    
    
    public void unbindPlayerConnector(){
    	dataTransmissionServlet.unbindPlayerConnector();
		core.log(Logger.DEBUG, this.getClass(), "Application Management Servlet unbind with PlayerConnector");	
    }
    
    
    protected PlayerConnector getPlayerConnector(){
    	return playerConnector;
    }
    
    
    public void bindHttpService(HttpService service){
    	try {    		
	    	// create a default context to share between registrations
	        HttpContext httpContext = httpService.createDefaultHttpContext();

	        // register the hello world servlet
	        Dictionary<String,String> initParams = new Hashtable<String,String>();
	        initParams.put( "from", "HttpService" );

	        httpService.registerServlet(SERVLET_URL, this, initParams, httpContext);	        
			core.log(Logger.DEBUG, this.getClass(), "Application Management Servlet registered to HTTP service on URL = "+SERVLET_URL);
		    
		    // instantiate dataTransmissionServlet
	        dataTransmissionServlet.bindHttpService(service);
		    
    	} catch (Throwable t) {
			core.log(Logger.WARN, this.getClass(), "Exception on binding of Application Management Servlet with HTTP service", t);
		}
    }
    
    
    public void unbindHttpService(HttpService httpService){
    	if(httpService != null){
    		dataTransmissionServlet.unbindHttpService(httpService);
        	httpService.unregister(SERVLET_URL);
        	if(core != null){
        		core.log(Logger.DEBUG, this.getClass(), "Application Management Servlet unregistered to HTTP service");
        	}
    	}else{
    		if(core != null){
				core.log(Logger.WARN, this.getClass(), "Application Management Servlet cannot unregistered to HTTP service since it has a null value");
    		}
    	}
    }
    
    
    public void doGet(HttpServletRequest request,HttpServletResponse response) throws IOException, ServletException {}
    
    
    public void doPost(HttpServletRequest request,HttpServletResponse response) throws IOException, ServletException {
    	int requesterType = Integer.parseInt(request.getParameter("from"));
    	
    	switch(requesterType){
    		case 0: // Admin request
    			doPostAdmin(request,response);
    			break;
    			
    		case 1: // GameMaster
    			break;
    			
     		case 2: // Player request
     			doPostPlayer(request,response);
     			break;
     			
     		case 3: // Sensor request
     			break;
    	}
    }
    
    
    private void doPostPlayer(HttpServletRequest request,HttpServletResponse response) throws IOException, ServletException {
		// get the "mode" parameter from request
		String mode = request.getParameter("mode");
		// switching mode
		if (mode.equals("FirstLogin")){
			doPostFirstLogin(request, response);
		}else if (mode.equals("Login")){
			doPostLogin(request, response);
		}else if (mode.equals("Lobby")){
			doPostLobby(request, response);
		}else if (mode.equals("Name")){
			doPostName(request, response);
		}else if (mode.equals("Join")){
			doPostJoin(request, response);
		}else if (mode.equals("Create")){
			doPostCreate(request, response);
		}else if (mode.equals("StartAI")){
			doPostStart(request, response);
		}else if (mode.equals("EndAI")){
			end(request, response);
		}else if (mode.equals("QuitAI")){
			quitApplicationInstance(request, response);
		}else if (mode.equals("Quit")){
			quitPlatform(request, response);
		}
	}
	
	
	private void doPostAdmin(HttpServletRequest request,HttpServletResponse response) {
		// for future distant management
	}	

		
	private void doPostFirstLogin(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters
		short userID = Short.parseShort(request.getParameter("uid"));
		short applicationID = Short.parseShort(((String) request.getParameter("appID")));
		String username = request.getParameter("username");
		String password = request.getParameter("password");
		core.log(Logger.TRACE, this.getClass(), "First Login request received from: userID = "+userID+", applicationID = "+applicationID+", username = "+username);
		// the actor id
		int actorID;
		// get a lock on the player connector
		synchronized(playerConnector) {
			// registration of the user(username,password) in uGASP database for a special application
			actorID = playerConnector.firstLogin(userID,applicationID,username,password);	
		}
		
		// Sending actorID in the response
		DataOutputStream dos = null;
		try{
			// get the response's data output stream 
			dos = new DataOutputStream(response.getOutputStream());
			// write the short actorID on the stream
			dos.writeShort(Short.parseShort(actorID+""));
			core.log(Logger.TRACE, this.getClass(), "First Login response to userID = "+userID+" : actorID = "+actorID);
		}catch(Throwable t){
			core.log(Logger.ERROR, this.getClass(), "Exception while responding on firstLogin ",t);
	    }finally{
	    	// trying to close the stream if necessary
	    	if (dos != null){
	    		try{
	    			dos.close();
	    		}catch (Throwable t) {
	    			core.log(Logger.WARN, this.getClass(), "Exception on closing of data output stream", t);
	    		}
	    	}
	    }
	}
	
	
	private void doPostLogin(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters 
		short actorID = Short.parseShort(request.getParameter("aID"));
		String username = request.getParameter("username");
		String password = request.getParameter("password");
		core.log(Logger.TRACE, this.getClass(), "Login request received from: actorID = "+actorID+", username = "+username);
		// the session id
		int sessionID;
		// get a lock on the player connector
		synchronized (playerConnector) {
			// connect to the application for which user has been registered in uGASP database
			sessionID = playerConnector.login(actorID, username, password);
		}
		
		// Sending sessionID in the response
		DataOutputStream dos = null;
		try{
			// get the response's data output stream 
			dos = new DataOutputStream(response.getOutputStream());
			// write the short sessionID on the stream
			dos.writeShort(Short.parseShort(sessionID+""));
			core.log(Logger.TRACE, this.getClass(), "Login response to actorID = "+actorID+" : sessionID = "+sessionID);
		}catch(Throwable t){
			core.log(Logger.ERROR, this.getClass(), "Exception while responding on login ",t);
	    }finally{
	    	// trying to close the stream if necessary
	    	if (dos != null){
	    		try{
	    			dos.close();
	    		}catch (Throwable t) {
	    			core.log(Logger.WARN, this.getClass(), "Exception on closing of data output stream", t);
	    		}
	    	}
	    }
	}
	
	
	private void doPostLobby(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters 
		short sessionID = Short.parseShort(request.getParameter("sID"));
		core.log(Logger.TRACE, this.getClass(), "Lobby request received from: sessionID = "+sessionID);
		// Sending list of available instance in the response
		DataOutputStream dos = null;
		try{
			// get the response's data output stream 
			dos = new DataOutputStream(response.getOutputStream());
			// get a lock on the player connector
			synchronized (playerConnector) {
				// get the list of available instances for the application the user has just logged in to
				playerConnector.lobby(sessionID,dos);
			}
		}catch(Throwable t){
			core.log(Logger.ERROR, this.getClass(), "Exception while responding on lobby ",t);
	    }finally{
	    	// trying to close the stream if necessary
	    	if (dos != null){
	    		try{
	    			dos.close();
	    		}catch (Throwable t) {
	    			core.log(Logger.WARN, this.getClass(), "Exception on closing of data output stream", t);
	    		}
	    	}
	    }
	}
	
	
	private void doPostName(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters
		short actorSessionID = Short.parseShort(request.getParameter("aSID"));
		String name = request.getParameter("name");
		core.log(Logger.TRACE, this.getClass(), "Name request received from: actorSessionID = "+actorSessionID+" name = "+name);
		// the pseudo
		String pseudo;
		// get a lock on the player connector
		synchronized (playerConnector) {
			// set the user pseudo
			pseudo = playerConnector.name(actorSessionID,name);
		}
		
		// Sending the pseudo in the response
		DataOutputStream dos = null;
		try{
			// get the response's data output stream 
			dos = new DataOutputStream(response.getOutputStream());
			// write the pseudo on the stream
			dos.writeUTF(pseudo);
			core.log(Logger.TRACE, this.getClass(), "Name request response to actorSessionID = "+actorSessionID+" : new pseudo = "+pseudo);
		}catch(Throwable t){
			core.log(Logger.ERROR, this.getClass(), "Exception while responding on lobby ",t);
	    }finally{
	    	// trying to close the stream if necessary
	    	if (dos != null){
	    		try{
	    			dos.close();
	    		}catch (Throwable t) {
	    			core.log(Logger.WARN, this.getClass(), "Exception on closing of data output stream", t);
	    		}
	    	}
	    }
	}
	
	
	private void doPostJoin(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters
		short sessionID =  Short.parseShort(request.getParameter("sID"));
		String aIID = request.getParameter("aIID");
		// if aIID is null, applicationInstanceID is set to 0
		short applicationInstanceID = aIID.equals("")?0:Short.parseShort(aIID);
		core.log(Logger.TRACE, this.getClass(), "Join Application Instance request received from sessionID = "+sessionID+" to join applicationInstance with ID = "+applicationInstanceID);
		short actorSessionID;
		
		// get a lock on the player connector
		synchronized (playerConnector) {
			// join the application identificated by its applicationInstanceID
			actorSessionID = playerConnector.joinAI(sessionID,applicationInstanceID);
		}
		
		// Sending actorSessionID in the response
		DataOutputStream dos = null;
		try{
			// get the response's data output stream 
			dos = new DataOutputStream(response.getOutputStream());
			// write the short actorSessionID on the stream
			dos.writeShort(actorSessionID);
			core.log(Logger.TRACE, this.getClass(), "Join Application Instance response to sessionID = "+sessionID+" : actorSessionID = "+actorSessionID);
		}catch(Throwable t){
			core.log(Logger.ERROR, this.getClass(), "Exception while responding on lobby ",t);
	    }finally{
	    	// trying to close the stream if necessary
	    	if (dos != null){
	    		try{
	    			dos.close();
	    		}catch (Throwable t) {
	    			core.log(Logger.WARN, this.getClass(), "Exception on closing of data output stream", t);
	    		}
	    	}
	    }
	}
	
	
	private void doPostCreate(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters 
		short sessionID =  Short.parseShort(request.getParameter("sID"));
		int minActors = Integer.parseInt(request.getParameter("minActors"));
		int maxActors = Integer.parseInt(request.getParameter("maxActors"));
		String[] actors = request.getParameterValues("actors");
		core.log(Logger.TRACE, this.getClass(), "Create Application Instance request received from sessionID = "+sessionID+" with min Actors = "+minActors+" and max Actors = "+maxActors);
		// the session id
		short actorSessionID;
		// get a lock on the player connector
		synchronized (playerConnector) {
			// create an application instance
			actorSessionID = playerConnector.create(sessionID, minActors, maxActors, actors);
		}
		
		// Sending actorSessionID in the response
		DataOutputStream dos = null;
		try{
			// get the response's data output stream 
			dos = new DataOutputStream(response.getOutputStream());
			// write the short actorSessionID on the stream
			dos.writeShort(actorSessionID);
			core.log(Logger.TRACE, this.getClass(), "Create Application Instance response to sessionID = "+sessionID+" : actorSessionID = "+actorSessionID);
		}catch(Throwable t){
			core.log(Logger.ERROR, this.getClass(), "Exception while responding on lobby ",t);
	    }finally{
	    	// trying to close the stream if necessary
	    	if (dos != null){
	    		try{
	    			dos.close();
	    		}catch (Throwable t) {
	    			core.log(Logger.WARN, this.getClass(), "Exception on closing of data output stream", t);
	    		}
	    	}
	    }
	}
	
	
	private void doPostStart(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters
		short actorSessionID =  Short.parseShort(request.getParameter("aSID"));
		core.log(Logger.TRACE, this.getClass(), "Start Application Instance request received from Actor Session with ID = "+actorSessionID);
		boolean applicationInstanceStarted;
		// get a lock on the player connector
		synchronized (playerConnector) {
			// start an application instance
			applicationInstanceStarted = playerConnector.start(actorSessionID);
		}
		
		// Sending aID in the response
		DataOutputStream dos = null;
		try{
			// get the response's data output stream 
			dos = new DataOutputStream(response.getOutputStream());
			// write the result of the start method on the stream
			dos.writeBoolean(applicationInstanceStarted);
			core.log(Logger.TRACE, this.getClass(), "Start Application Instance response to Actor Session with ID = "+actorSessionID+": "+applicationInstanceStarted);
		}catch(Throwable t){
			core.log(Logger.ERROR, this.getClass(), "Exception while responding on lobby ",t);
	    }finally{
	    	// trying to close the stream if necessary
	    	if (dos != null){
	    		try{
	    			dos.close();
	    		}catch (Throwable t) {
	    			core.log(Logger.WARN, this.getClass(), "Exception on closing of data output stream", t);
	    		}
	    	}
	    }
	}
	
	
	private void end(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters
		short actorSessionID = Short.parseShort(request.getParameter("aSID"));
		core.log(Logger.TRACE, this.getClass(), "End Application Instance request received from Actor Session with ID = "+actorSessionID);
		// get a lock on the player connector
		synchronized (playerConnector){
			// kill the application instance (in manual mode, only OWNER can perform end method)
			playerConnector.end(actorSessionID);
		}
	}
	
	
	private void quitApplicationInstance(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters
		short actorSessionID = Short.parseShort(request.getParameter("aSID"));
		core.log(Logger.TRACE, this.getClass(), "Quit Application Instance request received from Actor Session with ID = "+actorSessionID);
		// get a lock on the player connector
		synchronized (playerConnector){
			// user identified by its actorSessionID quits the application instance
			playerConnector.quitAI(actorSessionID);
		}
	}
	
	
	private void quitPlatform(HttpServletRequest request,HttpServletResponse response){
		// get http request parameters
		short sessionID = Short.parseShort(request.getParameter("sID"));
		core.log(Logger.TRACE, this.getClass(), "Quit Platform request received from Session with ID = "+sessionID);
		// get a lock on the player connector
		synchronized (playerConnector){
			// user identified by its sessionID quits the platform
			playerConnector.quit(sessionID);
		}
	}	
}