/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.impl.connector.http;


import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Dictionary;
import java.util.Hashtable;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.ugasp.connector.manager.DataTransmissionRequestManager;
import org.ugasp.connector.manager.impl.DataTransmissionRequestManager_v1;
import org.ugasp.core.Platform;
import org.ugasp.logging.Logger;
import org.ugasp.service.external.player.PlayerConnector;


/**
 * This servlet is used for data transmission between clients and platform:
 * <ul>
 * <li>clients send its Remote Method Asynchronous Invocation and its MooDS object</li>
 * <li>platform sends events to clients, if so</li>
 * </ul>
 */
public class DataTransmissionServlet extends HttpServlet{


	private static final byte HIGHEST_DATA_TRANSMISSION_PROTOCOL_VERSION = 1;
	private static final String SERVLET_URL = "/uGASP/DataTransmissionServlet";

	private HttpService 	httpService;
	
	private Platform		core;
	private PlayerConnector playerConnector;

	public DataTransmissionServlet(){}

	public void init(){ 
		log(Logger.DEBUG, "Data Transmission Servlet initialized");
	}

	public void start(){
		log(Logger.DEBUG, "Data Transmission Servlet started");
	}

	public void stop(){
		log(Logger.DEBUG, "Data Transmission Servlet stopped");
	}

	
	public void bindPlatform(Platform core){
		this.core = core;
		log(Logger.DEBUG, "Data Transmission Servlet bind with ugasp platform");
	}

	public void unbindPlatform(){
		log(Logger.DEBUG, "Data Transmission Servlet unbind with ugasp platform");
		core = null;
	}
	
	
	public void bindPlayerConnector(PlayerConnector connector){
		this.playerConnector = connector;
		log(Logger.DEBUG, "Data Transmission Servlet bind with PlayerConnector");
	}

	public void unbindPlayerConnector(){
		playerConnector = null;
		log(Logger.DEBUG, "Data Transmission Servlet unbind with PlayerConnector");	
	}


	public void bindHttpService(HttpService service){
		try {
			// get the httpService from bundle main servlet
			httpService = service;

			// create a default context to share between registrations
			HttpContext httpContext = httpService.createDefaultHttpContext();

			// register the hello world servlet
			Dictionary<String,String> initParams = new Hashtable<String,String>();
			initParams.put( "from", "HttpService" );

			httpService.registerServlet(SERVLET_URL, this, initParams, httpContext);
			log(Logger.DEBUG, "Data Transmission Servlet registered to HTTP service on URL = "+SERVLET_URL);
		} catch (Throwable t) {
			log(Logger.WARN, "Exception on binding of Data Transmission Servlet with HTTP service", t);
		}
	}

	public void unbindHttpService(HttpService httpService){
		if(httpService != null){
			httpService.unregister(SERVLET_URL);
			log(Logger.DEBUG, "Data Transmission Servlet unregistered to HTTP service");
		}else{
			log(Logger.WARN, "Data Transmission Servlet cannot unregistered to HTTP service since it has a null value");
		}
	}

	public void doGet(HttpServletRequest request,HttpServletResponse response) throws IOException, ServletException {}


	/***********************************************************************/
	/*************************** New management of protocols ***************/
	/***********************************************************************/

	
	/**
	 * This method is called on the reception of a client request. It successively:
	 * <ul>
	 * <li>retrieve Data Input Stream from the servlet request</li>
	 * <li>check whether the whole request has been received properly, reading the request size</li>
	 * <li>get the version of the data transmission protocol used by the client</li>
	 * <li>create a DataTransmissionRequestManager matching with this prtocol version, if so</li> 
	 * <li>pass a copy of the request input stream and the response output stream to the DataTransmissionRequestManager</li>
	 * <li>make the DataTransmissionRequestManager treat the data transmitted by client, and possibly send data to it, if so.<li>
	 * </ul>
	 * 
	 */
	public void doPost(HttpServletRequest request,HttpServletResponse response) throws IOException, ServletException {
		// instantiation of streams and arrays
		DataInputStream 		requestInputStream 				= null;
		ServletOutputStream 	responseOutputStream 			= null;
		ByteArrayOutputStream 	intermediate 					= new ByteArrayOutputStream();
		DataOutputStream		intermediateDataOutputStream 	= new DataOutputStream(intermediate);
		byte[] data;
		try{
			// retrieve data from the request
			data = retrieveDataFromRequest(request);
			if(data != null){
				// create a DataInputStream from data
				requestInputStream = new DataInputStream(new ByteArrayInputStream(data));
				// check the header of the request (message length and protocol version)        	
				requestInputStream = checkClientRequest(requestInputStream);
				// retrieve client version of data transmission protocol
				byte receivedProtocolVersion = requestInputStream.readByte();
				// test whether protocol version is the proper one
				if(receivedProtocolVersion > HIGHEST_DATA_TRANSMISSION_PROTOCOL_VERSION){
					log(Logger.WARN, 	"Bad data transmission protocol version received = "+
										receivedProtocolVersion+
										", highest version of protocol supported = "+
										HIGHEST_DATA_TRANSMISSION_PROTOCOL_VERSION);
					// send the highest version of data transmission protocol handled by the platform
					sendPlatformHighestProtocolVersion(intermediateDataOutputStream);
				}else{
					long start = System.currentTimeMillis();
					treatClientRequest(receivedProtocolVersion, requestInputStream, intermediateDataOutputStream);
					long end = System.currentTimeMillis();
					log(Logger.TRACE,"Client request treatment duration = "+(end-start)+" ms");
				}
				// if platform has data to send back to client
				if(intermediate.size() != 0){
					// retrieving response output stream
					responseOutputStream = response.getOutputStream();
					// writing data on it
					responseOutputStream.write(intermediate.toByteArray());
				}else{
					
				}
			}
		} catch(Throwable t){
			log(Logger.ERROR, "Exception occurred during data transmission", t);
		} finally{
			if (requestInputStream!=null) {
				try{
					requestInputStream.close();
				} catch (Throwable t){
					log(Logger.WARN,"Exception occured while trying to close request input stream",t);
				}
			}
			if (responseOutputStream!=null){
				try{
					responseOutputStream.close();
				} catch (Throwable t){
					log(Logger.WARN,"Exception occured while trying to close response output stream",t);
				}
			}
			intermediateDataOutputStream.close();
			intermediate.close();
		}
	}

	
	private void treatClientRequest(byte receivedProtocolVersion, DataInputStream requestInputStream, DataOutputStream responseOutputStream) throws IOException{
		DataTransmissionRequestManager requestManager = null;
		// switch treatment depending on client data transmission protocol version
		switch(receivedProtocolVersion){
		case DataTransmissionRequestManager_v1.DATA_TRANSMISSION_PROTOCOL_VERSION:
			requestManager = new DataTransmissionRequestManager_v1(requestInputStream, responseOutputStream, core, playerConnector);
			break;
		default:
			break;
		}
		if(requestManager != null){
			// treat client request
			requestManager.treatClientRequest();             		
		}else{
			log(Logger.WARN, 	"Data transmission protocol version received = "+
								receivedProtocolVersion+
								" is no more handled by this version of ugasp platform");
		}
	}
	

	/*
	 * Check whether message length sent by client matches with bytes available in the dataTransmissionInputStream.
	 */
	private DataInputStream checkClientRequest(DataInputStream dataTransmissionInputStream) throws IOException{
		if(dataTransmissionInputStream == null){
			log(Logger.WARN, "Cannot check client request since data transmission input stream has a null value");
			return null;
		}
		// retrieve the length of the message sent by uGASP
		int messageLength = dataTransmissionInputStream.readInt();
		if(messageLength != dataTransmissionInputStream.available()){
			log(Logger.WARN,"Client request is ignored since required message length ("+
							messageLength+
							") doesn't match with data transmission input stream available bits ("+
							dataTransmissionInputStream.available()+")");
			return null;
		}
		return dataTransmissionInputStream;
	}
	
	private byte[] retrieveDataFromRequest(HttpServletRequest request) throws IOException{
		byte[] data = null;
		if(request != null){
			BufferedInputStream requestBufferInputStream = new BufferedInputStream(request.getInputStream());
			int i = 0;
			int read;
			// initialize data
			data = new byte[request.getContentLength()];
			try {
				while((read = requestBufferInputStream.read()) != -1 ){
					data[i] = (byte)read;
					i++;
				}
			} catch (Throwable t) {
				log(Logger.WARN,"Exception on the retrieving of data from client's request",t);
				throw new IOException();
			}
			//printData(data);
		}else{
			log(Logger.WARN, "Client request has a null value");
		}
		return data;
		
	}
	
	
	/*
	 * Send the highest version of data transmission protocol handled by the platform
	 */
	private void sendPlatformHighestProtocolVersion(DataOutputStream responseOutputStream) throws IOException{
		// write message length: just 1 byte for protocol version
		responseOutputStream.writeInt(1);
		// highest protocol version handled by the platform
		responseOutputStream.writeByte(HIGHEST_DATA_TRANSMISSION_PROTOCOL_VERSION);
		
	}
	
	private void log(byte level, String message){
		if(core != null){
			core.log(level, getClass(), message);
		}else{
			// no log level with System.out
			System.out.println("Log message: "+message);
		}
	}
	
	private void log(byte level, String message, Throwable t){
		if(core != null){
			core.log(level, getClass(), message,t);
		}else{
			// no log level with System.out
			System.out.println("Log message: "+message);
			t.printStackTrace();
		}
	}
	
	private void printData(byte[] data){
		log(Logger.TRACE,"Data retrieved from client request (byte array)");
		for (int j = 0; j < data.length; j++) {
			log(Logger.TRACE,"--- request["+j+"] = "+data[j]);
		}
	}

}