/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.core;


import java.util.List;

import org.ugasp.core.utils.ApplicationInstanceInfos;

import org.ugasp.protocol.moods.CustomTypes;

import org.ugasp.eventmodel.event.MooDSEvent;
import org.ugasp.eventmodel.listener.OnEndEvent;
import org.ugasp.eventmodel.listener.OnJoinEvent;
import org.ugasp.eventmodel.listener.OnMooDSEvent;
import org.ugasp.eventmodel.listener.OnQuitEvent;
import org.ugasp.eventmodel.listener.OnStartEvent;
import org.ugasp.eventmodel.listener.OnRMAIEvent;


/**
 * ApplicationInstance represent a game session of a specific application.
 * It contains a vector of ActorSession representing the actors want to play in.
 * This classes manage the in game events.
 *
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project 
 * 
 */

public interface ApplicationInstance extends OnJoinEvent, OnStartEvent, OnEndEvent , OnQuitEvent, OnRMAIEvent, OnMooDSEvent{

	/**
	 * Returns the ID of the ApplicationInstance.
	 * 
	 * @return the ApplicationInstanceID
	 */
	int getApplicationInstanceID();
	
	/**
	 * Returns the ID of the Application.
	 * 
	 * @return the ApplicationID
	 */
	int getApplicationID();
	
	/**
	 * Returns the minimum number of actors required to start the game session.
	 * 
	 * @return the minimum number of actors
	 */
	int getMinActors();
	
	
	/**
	 * Returns the maximum number of actors can play the game session.
	 * 
	 * @return the maximum number of actors
	 */
	int getMaxActors();
	
	
	/**
	 * Returns the ActorSessionID of the ApplicationInstance owner.
	 * 
	 * @return the ActorSessionID
	 */
	short getOwnerActorSessionID();
	
	
	/**
	 * Returns the encoder of the application associated with the application instance.
	 * 
	 * @return the Actor ID
	 */
	CustomTypes getCustomTypes();
	
	
	/**
	 * Determine if the ApplicationInstance is public or not.
	 * 
	 * @return boolean 
	 */
	boolean isPublic();
	
	
	/**
	 * Determine if the ApplicationInstance is running or not.
	 * 
	 * @return boolean
	 */
	boolean isRunning();
	
	
	/**
	 * Determine if the Owner Master Application must destroy the application instance or not.
	 * 
	 * @return boolean
	 */
	boolean isDestroyable();
	
	/**
	 * Determine if the the application instance is joignable by a player or not.
	 * 
	 * @return boolean
	 */
	boolean isJoinable();
	
	
	/**
	 * Determine if the the application instance is startable by the player owner or not.
	 * 
	 * @return boolean
	 */
	boolean isStartable();
	
	
	/**
	 * Get an unmodifiable view of the list of actor sessions.
	 * @return
	 */
	List<ActorSession> getActorSessionsList();
	
	
	/**
	 * Return the pseudo of a user from its actorSessionId.
	 * 
	 * @return	the pseudo of user (Beware, pseudo is NOT the user name).
	 */
	String getPseudoByActorSessionId(short actorSessionId);
	
	
	 /**
	 * Return actorSessionId of a user from its pseudo.
	 * 
	 * @return	the actorSessionId of user
	 * 			-1 if no actorSessionId is matching with pseudo.
	 */
	short getActorSessionIdByPseudo(String pseudo);
	
	
	/**
	 * Create a new ActorSession in the ApplicationInstance corresponding to 
	 * the user Session object informations.
	 * Returns the ActorSession ID.
	 * 
	 * @param aIID the ApplicationInstance
	 * @param session the Session object of the user
	 * @return the ActorSession ID
	 */
	short createNewActorSession(int aIID, Session session);
	

	/**
	 * Returns the ActorSession object corresponding to the ActorSession ID.
	 * 
	 * @param actorSessionID the ActorSession ID
	 * @return the ActorSession object
	 */
	ActorSession getActorSession(short actorSessionID);

	
	/**
	 * Remove the ActorSession object corresponding to the ActorSessionID.
	 * 
	 * @param actorSessionID the ActorSesion ID
	 * @return boolean representing the operation success
	 */
	boolean removeActorSession(short actorSessionID);
	

	/**
	 * Determine if the ActorSession associated with the ActorSession ID is contained by
	 * this ApplicationInstance or not.
	 * 
	 * @param actorSessionID
	 * @return boolean "present or not"
	 */
	boolean containsActorSession(short actorSessionID);
	
	/**
	 * Retrieve the actorID associated with a actorSessionID.
	 * 
	 * @param actorSessionID
	 * @return boolean "present or not"
	 */
	int retrieveActorID(short actorSessionID);

	
	/**
	 * For Lobby service, this method provide informations of the ApplicationInstance.
	 * 
	 * @return an ApplicationInstanceInfos object
	 */
	ApplicationInstanceInfos getApplicationInstanceInfos();
	
	
	/**
	 * Specify the name of the Application Instance.
	 * 
	 * @param appInstanceName
	 */
	void setApplicationInstanceName(String appInstanceName);
	
	
	/**
	 * Add the actor to the actors vector.
	 * 
	 * @param actorSession the ActorSession object
	 * @return boolean of the operation success
	 */
	boolean addActorSession(ActorSession actorSession);

    
    
    /**
     * Send player info to the actor session ASID.
     * In case of mobile intempestive deconnection, the reconnected mobile
     * must know the players state of the AI.
     * 
     * @param aSID
     */
    void sendPlayersInfos(short actorSessionID);
    
    
    /**
	 * Send a MooDS Event to a specific actorSession.
	 * 
	 * @param actorSessionID
	 * @param moodsEvent
	 */
    void sendMooDSDataToActorSession(short actorSessionID, MooDSEvent moodsEvent);


    
    /**
     * If the pseudo already exist in the AI modify it, else no modifications.
     * 
     * @param name
     * @return String the checked pseudo
     */
    String treatPseudo(String name);

    
    /**
     *  Raise a JoinEvent to all other players containing the new pseudo
     * @param actorSessionID
     */
    void raisePseudoModification(short actorSessionID, String pseudoname);

    
    /**
     *  Retrieve the applicationID from the master application instance.
     * @param actorSessionID
     */
    int retrieveApplicationID();

    /**
     * Destroy game server instance.
     * @param actorSessionID
     */
	void destroyServerInstance();

}