/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.core;


import java.util.Iterator;

import org.ugasp.service.internal.system.db.DBManager;
import org.ugasp.service.internal.system.id.IDManager;


/**
 * Platform is the base of the GAming Services Platform (GASP), the container of all GP representing objects.
 * So it contains all MasterApplicationInstance objects in a vector.
 * It hold all of the service instances (Lobby,...) and system instances (DB manager, ID manager, ...).
 * 
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project 
 * 
 */

public interface Platform {

	/**
	 * Returns the MasterApplicationInstance associated with the application ID or if it is not
	 * already instantiates, initialize it.
	 * 
	 * @param applicationID the application ID
	 * @return the MasterApplicationInstance object
	 */
	MasterApplicationInstance getMasterApplicationInstance(int applicationID);


	/**
	 * Remove the MasterApplicationInstance associated with the application ID.
	 * 
	 * @param applicationID the application ID
	 * @return boolean of the operation success
	 */
	boolean removeMasterApplicationInstance(int applicationID);


	/**
	 * Determine if the vector of MasterApplicationInstance is empty or not.
	 * 
	 * @return boolean
	 */
	boolean isMasterApplicationInstanceEmpty();

	
	/**
	 * Return the number of MasterApplicationInstance currently instantiated.
	 * 
	 * @return the number of MasterApplication instances.
	 **/
	int masterApplicationInstanceSize();
	
	
	/**
	 * Returns an iterator on the MasterApplicationInstance vector.
	 * 
	 * @return iterator
	 */
	Iterator<MasterApplicationInstance> enumerateMasterApplicationInstance();


	/**
	 * Determine if the MasterApplicationInstance of the application associated with the
	 * application ID is initialized or not.
	 * 
	 * @param applicationID
	 * @return boolean
	 */
	boolean containsMasterApplicationInstance(int applicationID);

	
	/**
	 * Return the platform ID manager .
	 * 
	 * @return the IDManager object
	 */
	IDManager getIDManager();
	
	
	/**
	 * Return the platform DB manager .
	 * 
	 * @return the DBManager object
	 */
	DBManager getDBManager();
	
	
	/**
	 * Return the MasterApplicationInstance containing the sessionID.
	 * 
	 * @param sID the session ID
	 * @return the MasterApplicationInstance object
	 */
	public MasterApplicationInstance getMasterApplicationInstanceFromSessionID(int sessionID);
	
	
	/**
	 * Return the ApplicationInstance containing the actor session ID.
	 * 
	 * @param aSID the actor session ID
	 * @return the MasterApplicationInstance object
	 */
	public ApplicationInstance getApplicationInstanceFromActorSessionID(short actorSessionID);
	
	
	/**
	 * Return the MasterApplicationInstance containing the actor ID.
	 * 
	 * @param aID the actor ID
	 * @return the MasterApplicationInstance object
	 */
	public MasterApplicationInstance getMasterApplicationInstanceFromActorID(int actorID);
	
	
	/**
	 * Log a user in the platform, authentified by actorID, username and password.
	 * It consists of creating a new Session in the correct MasterApplicationInstance
	 * and return the session ID.
	 * 
	 * @param aID the Actor ID attributed at the first login
	 * @param username
	 * @param password
	 * 
	 * @return the SessionID of the session created or 0 if the authentification informations
	 * are not valid
	 */
	int login(int actorID, String username, String password);
	
	
	
	/**
	 * Provide to a user with a valid session ID to join a specific ApplicationInstance of
	 * the application open by the user. 
	 * Then create un ActorSession in the ApplicationInstance and return the ActorSessionID.
	 * 
	 * @param sID the Session ID
	 * @param aIID the ApplicationInstance ID
	 * @return the ActorSession ID
	 */
	short joinApplicationInstance(int sessionID, int applicationInstanceID);
	
	
	/**
	 * Provide to a user with a valid session ID to join a random ApplicationInstance
	 * corresponded to tbe application opened by the user.
	 * 
	 * @param sID the Session ID
	 * @return the ActorSession ID
	 */
	int joinRandomApplicationInstance(int sessionID);
	
	
	/**
	 * Provide to a user with a valid session ID to create an ApplicationInstance with a minimum
	 * start number of actors and a maximum number of actors.
	 * The ApplicationInstance status is:
	 * "public": if the table of actors is empty.
	 * "private": if the table of actors contains some Actor ID.
	 * Returns the ID of the ApplicationInstance created.
	 * 
	 * @param sID the Session ID
	 * @param minActors minimum number of actors for starting
	 * @param maxActors maximum number of actors
	 * @param actors a table of Actor ID
	 * @return ActorSessionID of the owner of the ApplicationInstance created.
	 */
	short createApplicationInstance(int sessionID, int minActors, int maxActors, String[] actors);
	
	
	/**
	 * This method is called when a user quit the application.
	 * ActorSession, Session, Actor instances corresponding to the user are removed.
	 * 
	 * @param sID the Session ID
	 */
	void quit(int sessionID);

	/**
	 * Return the platform timeout
	 */
	int getPlatformTimeout();
	
	
	public void log(byte logLevel, Class<?> clazz, String message);
	
	public void log(byte logLevel, Class<?> clazz, String message, Throwable t);
	
}


	
	
