/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.impl.core;


import java.util.Collections;
import java.util.List;
import java.util.Iterator;
import java.util.Vector;
import org.apache.felix.ipojo.InstanceManager;

import org.ugasp.core.Actor;
import org.ugasp.core.ActorSession;
import org.ugasp.core.ApplicationInstance;
import org.ugasp.core.utils.ApplicationInstanceInfos;
import org.ugasp.core.utils.ApplicationInstanceInfos.Player;
import org.ugasp.core.MasterApplicationInstance;
import org.ugasp.core.Session;
import org.ugasp.gamelogic.GameLogic;
import org.ugasp.logging.Logger;
import org.ugasp.protocol.moods.CustomTypes;
import org.ugasp.protocol.rmai.RMAIGameLogicManager.ActorSessionID;

import org.ugasp.eventmodel.event.*;
import org.ugasp.eventmodel.event.impl.*;
import org.ugasp.eventmodel.listener.*;


/**
 * ApplicationInstance represent a game session of a specific application.
 * It contains a vector of ActorSession representing the actors want to play in.
 * This classes manage the in game events.
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project 
 */
public class ApplicationInstanceImpl implements ApplicationInstance, OnJoinEvent, OnStartEvent, OnQuitEvent, OnEndEvent, OnRMAIEvent, OnMooDSEvent {


	private int applicationID;
	private int applicationInstanceID;
	private MasterApplicationInstance masterApp;
	private List<ActorSession> actorSessionsList;

	private short ownerActorSessionID;
	private int minA;
	private int maxA;
	private String applicationInstanceName; // for later implementation of game instance naming
	private boolean isPublicAI; // public or private  
	private volatile boolean isRunningAI; // running or waiting for players  
	private Vector<OnEvent> listeners;//Vector for events listener services: first listener to register is Server itself.
	private CustomTypes customTypes;
	private InstanceManager serverInstanceManager;
	private GameLogic gameLogic;

	private byte applicationModel;
	protected static final byte AUTO_APPLICATION_MODEL = 0;
	protected static final byte MANUAL_APPLICATION_MODEL = 1;
	protected static final byte PERMANENT_APPLICATION_MODEL = 2;
	protected static final byte MAPCYCLE_APPLICATION_MODEL = 3;

	protected static final short INVALID_APPLICATION_INSTANCE = -1;




	// this boolean is use to notify that the owner has left the AI
	// now the AI wait to all the actor sessions events vector in the 
	// AI will be empty
	private boolean toDestroy;

	private ApplicationInstanceImpl(){}	
	
	public static ApplicationInstanceImpl createApplicationInstanceImpl(MasterApplicationInstance masterApplication, Session ownerSession, int minActors, int maxActors, String[] actors, String path, byte model){
		ApplicationInstanceImpl appInstance = new ApplicationInstanceImpl();
		// link application instance with the suitable application
		appInstance.masterApp = masterApplication;
		appInstance.applicationID = appInstance.masterApp.getApplicationID();
		// create a game logic instance corresponding to the applicationID
		appInstance.serverInstanceManager = PlatformImpl.getPlatform().newGameLogic(appInstance.applicationID);
		if(appInstance.serverInstanceManager == null){
			appInstance = null;
		}else{
			// request for an ID
			appInstance.applicationInstanceID = PlatformImpl.getPlatform().requestAIID();
			PlatformImpl.getPlatform().log(Logger.INFO,ApplicationInstanceImpl.class,"Instance of Application number "+appInstance.applicationID+" instantiated with ID = "+appInstance.applicationInstanceID+", it needs at least "+minActors+" and at most "+maxActors+" actors");	
			//initialize vector of listeners
			appInstance.listeners = new Vector<OnEvent>();
			appInstance.gameLogic = (GameLogic) appInstance.serverInstanceManager.getPojoObject();
			appInstance.gameLogic.setOwnerAI(appInstance);
			appInstance.listeners.addElement(appInstance.gameLogic);
			//initialize other attributes
			appInstance.actorSessionsList = new Vector<ActorSession>();
			appInstance.minA = minActors;
			appInstance.maxA = maxActors;
			appInstance.applicationModel = model;
			if (actors.length == 0){
				appInstance.isPublicAI = true;
			}else{
				appInstance.addActors(actors);
				appInstance.isPublicAI = false;
			}
			appInstance.isRunningAI = false;
			appInstance.toDestroy = false;

			appInstance.ownerActorSessionID = appInstance.createNewActorSession(appInstance.applicationInstanceID,ownerSession);
			PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"Application Instance with ID "+appInstance.applicationInstanceID+" is owned by ActorSession with ID = "+appInstance.ownerActorSessionID);
		}
		return appInstance;
	}


	/**
	 * Returns the GameLogic associated with the ApplicationInstance.
	 * 
	 * @return the game logic
	 */
	public GameLogic getGameLogic() {
		return gameLogic;
	}


	/**
	 * Returns the ID of the ApplicationInstance.
	 * 
	 * @return the ApplicationInstanceID
	 */
	public int getApplicationInstanceID() {
		return applicationInstanceID;
	}

	/**
	 * Returns the ID of the Application.
	 * 
	 * @return the ApplicationID
	 */
	public int getApplicationID() {
		return applicationID;
	}

	/**
	 * Returns the minimum number of actors required to start the game session.
	 * 
	 * @return the minimum number of actors
	 */
	public int getMinActors(){return minA;}


	/**
	 * Returns the maximum number of actors can play the game session.
	 * 
	 * @return the maximum number of actors
	 */
	public int getMaxActors(){return maxA;}


	/**
	 * Returns the ActorId of the ApplicationInstanceID owner.
	 * 
	 * @return the ActorSession ID
	 */
	public short getOwnerActorSessionID(){return ownerActorSessionID;}


	/**
	 * This method return the Custom Types class matching with this application instance.
	 * If your application isn't designed for MooDS, an Exception will be thrown.
	 */
	public CustomTypes getCustomTypes() {
		if(customTypes == null){
			customTypes = gameLogic.getCustomTypes();
		}
		return customTypes;
	}	


	/**
	 * Determine if the ApplicationInstance is public or not.
	 * 
	 * @return boolean 
	 */
	public boolean isPublic(){
		return isPublicAI;
	}


	/**
	 * Determine if the ApplicationInstance is running or not.
	 * 
	 * @return boolean
	 */
	public boolean isRunning() {
		return isRunningAI;
	}


	/**
	 * Determine if the owner master application can destroy or not the Application Instance.
	 */
	public boolean isDestroyable() {
		return toDestroy;
	}


	/**
	 * Determine if the the application instance is joignable by a player or not,
	 * eg if:
	 * the Application Instance is not destroyable
	 * the Application Instance is not running
	 * the max of players is reached
	 * 
	 * @return boolean
	 */
	public boolean isJoinable() {
		switch (applicationModel){
		case AUTO_APPLICATION_MODEL:
		case PERMANENT_APPLICATION_MODEL:	
			return (!this.isDestroyable()&&(this.actorSessionsList.size() < this.getMaxActors()));

		case MANUAL_APPLICATION_MODEL:
			return (!this.isDestroyable()
					&&!this.isRunning()
					&&(this.actorSessionsList.size() < this.getMaxActors()));
		case MAPCYCLE_APPLICATION_MODEL:
			return false;
		default: 
			PlatformImpl.getPlatform().log(Logger.WARN,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationInstanceID+" has a bad model number, you should check your model property in your metadata.xml (auto, manual or permanent)");
			return false;
		}	
	}


	/**
	 * Determine if the the application instance is startable by the player owner or not.
	 * 
	 * @return boolean
	 */
	public boolean isStartable() {
		if (this.getMinActors()<= this.actorSessionsList.size()) return true;
		return false;
	}


	/**
	 * Get an unmodifiable view of the list of actor sessions.
	 * @return
	 */
	public List<ActorSession> getActorSessionsList(){
		return Collections.unmodifiableList(actorSessionsList);
	}



	/**
	 * Create a new ActorSession in the ApplicationInstance corresponding to 
	 * the user Session object informations.
	 * Returns the ActorSession ID.
	 * 
	 * @param aIID the ApplicationInstance
	 * @param session the Session object of the user
	 * @return the ActorSession ID
	 */
	public short createNewActorSession(int aIID, Session session) {
		ActorSession actorSession = new ActorSessionImpl(aIID,session);
		PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"ActorSession with ID =  "+actorSession.getActorSessionID()+" is created by ApplicationInstance with ID = "+applicationInstanceID);

		actorSession.setPseudoName(masterApp.getActor(session.getActorID()).getPseudoName()); 
		this.addActorSession(actorSession);

		JoinEvent joinEv = new JoinEventImpl(actorSession.getActorSessionID(), actorSession.getPseudoName());
		onJoinEvent(joinEv);
		switch (applicationModel){
		case AUTO_APPLICATION_MODEL : 
		case PERMANENT_APPLICATION_MODEL :
			PlatformImpl.getPlatform().log(Logger.INFO,ApplicationInstanceImpl.class,"Application Instance model set to "+MasterApplicationInstanceImpl.applicationModelToString(applicationModel)+", event is send by Platform to start instance");
			//when a player join the application instance, in this model, the game session automatically start
			StartEvent startEv = new StartEventImpl(ActorSessionID.uGASP_GAME_LOGIC.getID());
			if (isRunning())
				actorSession.raiseEvent(startEv);
			else onStartEvent(startEv);
			break;

		case MANUAL_APPLICATION_MODEL:
		case MAPCYCLE_APPLICATION_MODEL:
		default: break;
		}
		sendPlayersInfos(actorSession.getActorSessionID());
		return actorSession.getActorSessionID();
	}




	/**
	 * This method provide the new player to know all of the players currently
	 * in the application instance.
	 * @param actorSession the actor session associated to the player
	 */
	public void sendPlayersInfos(short actorSessionID) {
		ActorSession myActorSession = getActorSession(actorSessionID);
		PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"Application Instance("+applicationInstanceID+"): send to Actor Session("+myActorSession.getActorSessionID()+") the current list of players");

		// lock on the list of actor sessions
		synchronized (actorSessionsList) {
			for (ActorSession actorSessionRegistered : actorSessionsList) {
				// Create a join event for every actor session already registered
				JoinEventImpl joinEvent = new JoinEventImpl(actorSessionRegistered.getActorSessionID(),masterApp.getActor(actorSessionRegistered.getActorID()).getPseudoName());
				PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"Application Instance("+applicationInstanceID+"): raise event to actor session ("+actorSessionRegistered.getActorSessionID()+")");
				// send event to myActorSession
				myActorSession.raiseEvent(joinEvent);
			}
		}
	}


	/**
	 * Returns the ActorSession object corresponding to the ActorSession ID.
	 * 
	 * @param actorSessionID the ActorSession ID
	 * @return the ActorSession object
	 */
	public ActorSession getActorSession(short actorSessionID) {
		// lock on list of actor sessions
		synchronized (actorSessionsList) {
			for(ActorSession actorSession : actorSessionsList){
				if (actorSession.getActorSessionID() == actorSessionID){
					return actorSession;
				}
			}
		}
		return null;
	}


	/**
	 * Remove the ActorSession object corresponding to the ActorSessionID.
	 * 
	 * @param 	actorSessionID the ActorSesion ID
	 * @return 	true if actorSession is found and removed,
	 * 			false otherwise
	 */
	public boolean removeActorSession(short actorSessionID) {
		// lock on the actor sessions' list
		synchronized (actorSessionsList) {
			for (int i = 0; i < actorSessionsList.size(); i++) {
				// get actorSessionID for each actorSession in the list
				short actorSessionIDRegistered = actorSessionsList.get(i).getActorSessionID();
				// test if required actorSession is found
				if (actorSessionIDRegistered == actorSessionID) {
					// release actorSessionID for future actorSession
					PlatformImpl.getPlatform().releaseASID(actorSessionID);
					// remove actor session from the list
					actorSessionsList.remove(i);
					return true;
				}
			}
		}
		// otherwise
		return false;
	}


	/**
	 * Determine if the ActorSession associated with the ActorSession ID is contained by
	 * this ApplicationInstance or not.
	 * 
	 * @param actorSessionID
	 * @return boolean "present or not"
	 */
	public boolean containsActorSession(short actorSessionID) {
		// lock on the vector of actor sessions
		synchronized (actorSessionsList) {
			for(ActorSession actorSession : actorSessionsList){
				if (actorSession.getActorSessionID() == actorSessionID){
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * For Lobby service, this method provide informations of the ApplicationInstance.
	 * 
	 * @return an ApplicationInstanceInfos object
	 */
	public ApplicationInstanceInfos getApplicationInstanceInfos() {
		Vector<Player> playerList = new Vector<Player>();
		int nbPlayer;
		// lock on list of actor sessions
		synchronized (actorSessionsList) {
			nbPlayer = actorSessionsList.size();
			for (ActorSession actorSession : actorSessionsList){
				Player player = Player.getInstance(actorSession.getPseudoName(), actorSession.getActorID());
				playerList.add(player);
			}
		}

		ApplicationInstanceInfos appInstanceInfos = ApplicationInstanceInfos.getInstance(	this.getApplicationInstanceID(),
																							minA,
																							maxA,
																							nbPlayer,
																							playerList,
																							gameLogic.getOptionalInstanceInfos()
																						);

		return appInstanceInfos;
	}


	/**
	 * Returns the name of the ApplicationInstance.
	 * 
	 * @return the name of the Application Instance
	 */
	public String getApplicationInstanceName() {
		return applicationInstanceName;
	}


	/**
	 * Set the pool of actors required for a private Application
	 * Instance.
	 * 
	 * @param actors the table of actorID
	 */
	private void addActors(String[] actors) {
		/** TODO: for private Application Instance, to implement **/
	}


	/**
	 * Specify the name of the Application Instance.
	 * 
	 * @param appInstanceName
	 */
	public void setApplicationInstanceName(String appInstanceName) {
		applicationInstanceName = appInstanceName;
	}


	/**
	 * Add the actor to the actors vector.
	 * 
	 * @param actorSession the ActorSession object
	 * @return boolean of the operation success
	 */
	public boolean addActorSession(ActorSession actorSession){
		if (containsActorSession(actorSession.getActorSessionID()))
			return false;
		else{
			// lock on the vector of actor sessions
			synchronized (actorSessionsList) {
				actorSessionsList.add(actorSession);	
			}
			return true;
		}
	}



	/*********************************************************/
	/** 	In game Events management methods		 	    **/
	/**     & implementation of the events super classes     **/
	/*********************************************************/


	/**
	 * When a JoinEvent arrives, notify to all listeners and to all other actor sessions
	 */
	public void onJoinEvent(JoinEvent e) {
		if (!toDestroy){
			PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"JoinEvent received by Application Instance with ID = "+applicationID+" from ActorSession with ID = "+e.getActorSessionID());
			notifyJoinToListeners(e); 
			PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify JoinEvent to all OnEvent listeners");
			notifyToAllOtherActorSession(e);
			PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify JoinEvent to all Actor Sessions");
		}

	}


	/**
	 * When a StartEvent arrives, if the actor session ID of the event is
	 * the ownerActorSessionID then notify to all listeners and to all other actor sessions
	 * else do nothing.
	 */
	public void onStartEvent(StartEvent e) {
		if (!toDestroy){
			switch (applicationModel){
			case AUTO_APPLICATION_MODEL:
			case PERMANENT_APPLICATION_MODEL:
				PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"StartEvent received by Application Instance with ID = "+applicationID+" from Platform");
				isRunningAI = true;
				notifyStartToListeners(e);
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify StartEvent to all OnEvent listeners");
				notifyToAllOtherActorSession(e);
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify StartEvent to all Actor Sessions");
				break;

			case MANUAL_APPLICATION_MODEL:
				PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"StartEvent received by Application Instance with ID = "+applicationID+" from ActorSession with ID = "+e.getActorSessionID());
				if(e.getActorSessionID() == ownerActorSessionID){
					isRunningAI = true;
					notifyStartToListeners(e);
					PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify StartEvent to all OnEvent listeners");
					notifyToAllOtherActorSession(e);
					PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify StartEvent to all Actor Sessions");
				}
				else{
					PlatformImpl.getPlatform().log(Logger.WARN,ApplicationInstanceImpl.class,"StartEvent received by Application Instance with ID = "+applicationID+" from an ActorSession with ID = "+e.getActorSessionID()+" which isn't the owner (owner is ActorSession with ID = "+ownerActorSessionID+"), Instance couldn't start");
				}
				break;

			case MAPCYCLE_APPLICATION_MODEL:
				break;
			default: break;
			}   
		}
	}

	/**
	 * When a EndEvent arrives, if the actor session ID of the event is
	 * the ownerActorSessionID then notify to all listeners and to all other actor sessions
	 * else do nothing.
	 */
	public void onEndEvent(EndEvent e){
		if (!toDestroy){
			switch (applicationModel){
			case AUTO_APPLICATION_MODEL:
				PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"EndEvent received by Application Instance with ID = "+applicationID+" from ActorSession with ID = "+e.getActorSessionID());
				notifyEndToListeners(e);
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify EndEvent to all OnEvent listeners");
				notifyToAllOtherActorSession(e);	
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify EndEvent to all Actor Sessions");
				// how can instance be available if it doesn't run any more ??
				isRunningAI = false;
				//FOR DEBUG
				toDestroy = true;
				break;
			case PERMANENT_APPLICATION_MODEL:
				PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"EndEvent received by Application Instance with ID = "+applicationID+" from ActorSession with ID = "+e.getActorSessionID());
				notifyEndToListeners(e);
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify EndEvent to all OnEvent listeners");
				notifyToAllOtherActorSession(e); 
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify EndEvent to all Actor Sessions");
				isRunningAI = false;
				// Change value of toDestroy in ordeW. 
				// This way, when a client will call getEvents, it will raise a quit event.
				// when instance doesn't have clients anymore, it could be killed properly 
				toDestroy = true;
				break;
			case MANUAL_APPLICATION_MODEL:
				PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"EndEvent received by Application Instance with ID = "+applicationID+" from ActorSession with ID = "+e.getActorSessionID());
				if(e.getActorSessionID()==ownerActorSessionID){
					notifyEndToListeners(e);
					PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify EndEvent to all OnEvent listeners");
					notifyToAllOtherActorSession(e);
					PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify EndEvent to all Actor Sessions");
					isRunningAI = false;
					//FOR DEBUG
					toDestroy = true;
				}
				else{
					PlatformImpl.getPlatform().log(Logger.WARN,ApplicationInstanceImpl.class,"EndEvent received by Application Instance with ID = "+applicationID+" from an ActorSession with ID = "+e.getActorSessionID()+" which isn't the owner (owner is ActorSession with ID = "+ownerActorSessionID+"), Instance couldn't end");
				}
				break;
			case MAPCYCLE_APPLICATION_MODEL:
			default: break;
			}
		} 
	}


	/**
	 * When a QuitEvent arrives, then notify to all listeners and to all other actor sessions
	 */
	public void onQuitEvent(QuitEvent qe) {
		PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"QuitEvent received by Application Instance with ID = "+applicationID+" from ActorSession with ID = "+qe.getActorSessionID());
		removeActorSession(qe.getActorSessionID());
		masterApp.getSessionFromActorSessionID(qe.getActorSessionID()).setActorSessionID((short)0);
		if(!toDestroy){  
			notifyQuitToListeners(qe);
			PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify QuitEvent to all OnEvent listeners");
			notifyToAllOtherActorSession(qe);
			PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify QuitEvent to all Actor Sessions");
			switch (applicationModel){
			case AUTO_APPLICATION_MODEL :
				if (actorSessionsList.isEmpty()){
					toDestroy = true;
					PlatformImpl.getPlatform().log(Logger.INFO,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationInstanceID+" is ended");

					masterApp.removeApplicationInstance(applicationInstanceID); 
				}
				break;
			case MANUAL_APPLICATION_MODEL:
				if (qe.getActorSessionID() == this.ownerActorSessionID){
					// send a EndEvent to other players
					for (ActorSession actorSession : actorSessionsList) {
						if(actorSession.getActorSessionID() != this.ownerActorSessionID){
							actorSession.raiseEvent(new EndEventImpl(this.ownerActorSessionID));
						}
					}
					toDestroy = true;
					PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationInstanceID+" moved in destroyable mode");
				}
				if (actorSessionsList.isEmpty()){
					toDestroy = true;
					PlatformImpl.getPlatform().log(Logger.INFO,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationInstanceID+" is ended");
					masterApp.removeApplicationInstance(applicationInstanceID); 
				}
				break;
			case PERMANENT_APPLICATION_MODEL: // nothing is doing on quit event
				break;
			case MAPCYCLE_APPLICATION_MODEL: 
				break;
			}
		}
	}


	/**
	 * When a RMAIEvent arrives, notify to the suitable recipient, depending on the recipient
	 * actor session ID contained in the RMAIEvent.
	 */
	public void onRMAIEvent(RMAIEvent rmaiEvent) {
		if (!toDestroy){
			PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"RMAIEvent received from ActorSession with ID = "+rmaiEvent.getActorSessionID());
			// get recipient actor session ID
			short recipientActorSessionID = rmaiEvent.getRecipientActorSessionID();
			if (recipientActorSessionID > 0){
				// method should be invoked on a uGASP client (actor session ID > 0)
				// get suitable recipient
				ActorSession recipient = getActorSession(recipientActorSessionID);
				if(recipient != null){
					// raise rmai event on this recipient
					recipient.raiseEvent(rmaiEvent);
					PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"RMAIEvent sent to ActorSession with ID = "+recipientActorSessionID);
				}else{
					PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"No ActorSession could be retrieved with the matching ID = "+recipientActorSessionID);
				}
				
			}else if (recipientActorSessionID == ActorSessionID.uGASP_GAME_LOGIC.getID()){
				// method should be invoked only on uGASP gamelogic
				notifyRMAIToServer(rmaiEvent);
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"RMAIEvent sent to game logic");
			}else if (recipientActorSessionID == ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID()){
				// method should be invoked on a every uGASP client 
				notifyToAllOtherActorSession(rmaiEvent);
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"RMAIEvent broadcast to clients");
			}else if (recipientActorSessionID == ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID()){
				// method should be invoked on a every uGASP client ...
				notifyToAllOtherActorSession(rmaiEvent);
				// and on uGASP game logic
				notifyRMAIToServer(rmaiEvent);
				PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"RMAIEvent broadcast to clients and game logic");
			}
		}	        
	}


	/**
	 * When a MooDSEvent arrives, notify to all listeners.
	 */
	public void onMooDSEvent(MooDSEvent e) {
		if (!toDestroy){
			PlatformImpl.getPlatform().log(Logger.DEBUG,ApplicationInstanceImpl.class,"MooDSEvent received from ActorSession with ID = "+e.getActorSessionID());
			notifyMooDSDataToListeners(e);
			PlatformImpl.getPlatform().log(Logger.TRACE,ApplicationInstanceImpl.class,"Application Instance with ID = "+applicationID+" notify MooDSEvent to all OnEvent listeners");

		}
	}



	/**
	 * Provides to add a listener on events of application instance.
	 * 
	 * @param listener the object to become a listener
	 */
	public void addListener(OnEvent listener){
		listeners.add(listener);
	}


	/**
	 * Put the event in event vectors of the ApplicationInstance 
	 * actorSessionsList.
	 * 
	 * @param e the event
	 */
	private void notifyToAllOtherActorSession(Event e) {
		// lock on the list of actor sessions
		synchronized (actorSessionsList) {
			for(ActorSession actorSession : actorSessionsList){
				if (actorSession.getActorSessionID()!= e.getActorSessionID()) 
					actorSession.raiseEvent(e); 
			}	
		}
	}


	/**
	 * Raise the JoinEvent to all external listeners, in particular 
	 * the game server logic thread. 
	 * 
	 * @param e the JoinEvent
	 */
	private void notifyJoinToListeners(JoinEvent e) {
		for(Iterator<OnEvent> iter = listeners.iterator(); iter.hasNext();){
			OnJoinEvent listenner = (OnJoinEvent) iter.next();
			listenner.onJoinEvent(e);
		}
	}


	/**
	 * Raise the EndEvent to all external listeners, in particular 
	 * the game server logic thread.  
	 * 
	 * @param e the EndEvent
	 */
	private void notifyEndToListeners(EndEvent e) {
		for(Iterator<OnEvent> iter = listeners.iterator(); iter.hasNext();){
			OnEndEvent listenner = (OnEndEvent) iter.next();
			listenner.onEndEvent(e);
		}
	}


	/**
	 * Raise the StartEvent to all external listeners, in particular 
	 * the game server logic thread.  
	 * 
	 * @param e the StartEvent
	 */
	private void notifyStartToListeners(StartEvent e) {
		for(Iterator<OnEvent> iter = listeners.iterator(); iter.hasNext();){
			OnStartEvent listenner = (OnStartEvent) iter.next();
			listenner.onStartEvent(e);
		}

	}


	/**
	 * Raise the QuitEvent to all external listeners, in particular
	 * the game server logic thread. 
	 * 
	 * @param e the QuitEvent
	 */
	private void notifyQuitToListeners(QuitEvent e) {
		for(Iterator<OnEvent> iter = listeners.iterator(); iter.hasNext();){
			OnQuitEvent listenner = (OnQuitEvent) iter.next();
			listenner.onQuitEvent(e);
		}

	}


	// This method is used to notify uGASP server of an RMAI event.
	// Here, we don't want to notify all the listeners, cause only GameLogic shall
	// receive event for invoking method on itself.
	// TODO Should we throw an exception if listeners.get(0) isn't instance of GameLogic?
	// Or should we just improve addListeners documentation ?
	private boolean notifyRMAIToServer(RMAIEvent rmaiEvent){
		if(listeners.get(0) instanceof GameLogic){
			((OnRMAIEvent)(listeners.get(0))).onRMAIEvent(rmaiEvent);
			return true;
		} else return false;
	}


	/**
	 * Raise the DataEvent to all external listeners, in particular
	 * the game server logic thread. 
	 * 
	 * @param e the DataEvent
	 */
	private void notifyMooDSDataToListeners(MooDSEvent e) {
		for(Iterator<OnEvent> iter = listeners.iterator(); iter.hasNext();){
			OnMooDSEvent listenner = (OnMooDSEvent) iter.next();
			listenner.onMooDSEvent(e);
		}
	}


	/**
	 * Send a MooDS Event to a specific actorSession.
	 * 
	 * @param actorSessionID
	 * @param moodsEvent
	 */
	public void sendMooDSDataToActorSession(short actorSessionID, MooDSEvent moodsEvent) {
		if (!toDestroy){
			ActorSession actorSession = this.getActorSession(actorSessionID);
			actorSession.raiseEvent(moodsEvent);	
		}
	}    

	/**
	 * If the pseudo already exist in the AI modify it, else no modifications.
	 * 
	 * @param name
	 * @return valid pseudo name 
	 */
	public String treatPseudo(String name) {
		int i = 0;      
		// lock on the list of actor sessions
		synchronized (actorSessionsList) {
			for(ActorSession actorSession : actorSessionsList){
				if (actorSession.getPseudoName() == name) i++;
			}	
		}

		if (i > 0) name+= "("+i+")";
		return name;
	}

	/**
	 *  Raise a JoinEvent to all other players containing the new pseudo
	 * @param actorSessionID
	 */
	public void raisePseudoModification(short actorSessionID, String pseudoname) {
		JoinEvent je = new JoinEventImpl(actorSessionID, pseudoname);
		onJoinEvent(je);
	}

	/**
	 *  Retrieve the applicationID from the master application instance.
	 * @param actorSessionID
	 */
	public int retrieveApplicationID(){
		return masterApp.getApplicationID();
	}


	/**
	 * Destroy game server instance.
	 */
	public void destroyServerInstance() {
		serverInstanceManager.dispose();
		PlatformImpl.getPlatform().log(Logger.INFO,ApplicationInstanceImpl.class,"Server Instance Manager destroyed with Application Instance with ID = "+applicationInstanceID);
	}


	public int retrieveActorID(short actorSessionID) {
		ActorSession as = getActorSession(actorSessionID);
		if (as == null){
			return -1;
		}else{
			return as.getActorID();
		}
	}
	
	/**
	 * Return the pseudo of a user from its actorSessionId.
	 * 
	 * @return	the pseudo of user (Beware, pseudo is NOT the user name).
	 */
	public String getPseudoByActorSessionId(short actorSessionId){
		ActorSession actorSession = getActorSession(actorSessionId);
		Actor actor = masterApp.getActor(actorSession.getActorID());
		return actor.getPseudoName();
	}
	
	 /**
	 * Return actorSessionId of a user from its pseudo.
	 * 
	 * @return	the actorSessionId of user
	 * 			-1 if no actorSessionId is matching with pseudo.
	 */
	public short getActorSessionIdByPseudo(String pseudo){
		short actorSessionId = -1;
		synchronized (actorSessionsList) {
			for (ActorSession actorSession : actorSessionsList) {
				Actor actor = masterApp.getActor(actorSession.getActorID());
				if(actor.getPseudoName().equals(pseudo)){
					actorSessionId = actorSession.getActorSessionID();
					break;
				}
			}
		}
		return actorSessionId;
	}


	/***************************************************************************************************/
	/******************************  Methods inherited from Object *************************************/
	/***************************************************************************************************/

	public boolean equals(Object o){
		if(o == this){
			return true;
		}
		if(!(o instanceof ApplicationInstanceImpl)){
			return false;
		}
		ApplicationInstanceImpl applicationInstance = (ApplicationInstanceImpl)o;
		return  applicationInstance.applicationID == applicationID &&
		applicationInstance.applicationInstanceID == applicationInstanceID &&
		applicationInstance.applicationInstanceName.equals(applicationInstanceName) &&
		applicationInstance.masterApp.equals(masterApp) &&
		applicationInstance.ownerActorSessionID == ownerActorSessionID &&
		applicationInstance.minA == minA &&
		applicationInstance.maxA == maxA &&
		applicationInstance.isPublicAI == isPublicAI;
	}


	public int hashCode(){
		// Values 37 & 17 are from "The effective Java" by Joshua Bloch
		int result = 17;
		result = 37 * result + applicationID;
		result = 37 * result + applicationInstanceID;
		result = 37 * result + applicationInstanceName.hashCode();
		result = 37 * result + masterApp.hashCode();
		result = 37 * result + ownerActorSessionID;
		result = 37 * result + minA;
		result = 37 * result + maxA;
		result = 37 * result + Boolean.valueOf(isPublicAI).hashCode();
		return result;
	}
}
