/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.impl.core;


import java.util.List;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Collections;

import org.ugasp.core.Actor;
import org.ugasp.core.ApplicationInstance;
import org.ugasp.core.MasterApplicationInstance;
import org.ugasp.core.Session;
import org.ugasp.eventmodel.event.impl.QuitEventImpl;
import org.ugasp.logging.Logger;


/**
 * MasterApplicationInstance represents the container of all ApplicationInstance of a specific application.
 * So a MasterApplicationInstance contains a vector of ApplicationInstance.
 * It contains also a vector of Session objects, representing the Actors currently using the Application.
 * Finally it contains a vector of the Actor objects, representing a link User-Application.
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project 
 */


public class MasterApplicationInstanceImpl implements MasterApplicationInstance,Runnable {
	
	
	private int 	applicationID;
	private int 	masterApplicationID;
	private byte 	applicationModel;
	private String 	applicationPath;
	
	private int 	platformTimeout;
	private int 	applicationTimeout;
	
	private static final int 	DEFAULT_APPLICATION_TIMEOUT = 120000;
	private static final int 	PLAYERTIMEOUT_CHECKING_DELTATIME = 30000;
	
	public static final String 	AUTO_APPLICATION_MODEL 		= "auto";// starts, stops watching the number of actors
	public static final String 	MANUAL_APPLICATION_MODEL 	= "manual";// the owner of the aI starts, stops the aI
	public static final String	PERMANENT_APPLICATION_MODEL = "permanent";// session is not destroyed untill a client invoke endApplicationInstance().
	public static final String 	MAPCYCLE_APPLICATION_MODEL 	= "mapcycle";// unused yet 
	private static final byte 	DEFAULT_APPLICATION_MODEL 	= ApplicationInstanceImpl.AUTO_APPLICATION_MODEL;// default model is auto
	
	private List<ApplicationInstance> applicationInstancesList;
	private List<Session> sessionsList;
	private List<Actor> actorsList;
	
	private Thread timeoutThread = new Thread(this);
	
	
	public MasterApplicationInstanceImpl(int applicationID, int platformTimeout){
		PlatformImpl.getPlatform().log(Logger.INFO, MasterApplicationInstanceImpl.class,"Master Application Instance instantiation for application with ID = "+applicationID);
		this.platformTimeout	= platformTimeout;
		this.applicationID 		= applicationID;
		// Lists initialization
	    applicationInstancesList= new Vector<ApplicationInstance>();
		sessionsList 			= new Vector<Session>();
		actorsList 				= new Vector<Actor>();
		
		// request a masterApplicationID
		masterApplicationID = PlatformImpl.getPlatform().requestMAID();
		
		// request application properties
		Hashtable<?,?> appInfos = PlatformImpl.getPlatform().getApplicationProperties(applicationID);
		
		// set the application timeout
		Object value = appInfos.get("timeout");
		if (value != null){
			applicationTimeout=Integer.parseInt((String)value);
			PlatformImpl.getPlatform().log(Logger.INFO,MasterApplicationInstanceImpl.class, "Timeout for application with ID = "+applicationID+" is set to "+applicationTimeout+" ms");
		}else{
			applicationTimeout = DEFAULT_APPLICATION_TIMEOUT;
			PlatformImpl.getPlatform().log(Logger.INFO,MasterApplicationInstanceImpl.class,"No application timeout specified for application with ID = "+applicationID+", timeout set to default value: "+DEFAULT_APPLICATION_TIMEOUT+" ms");
		}
		// set the application model
		value = appInfos.get("model");
		if (value != null){
			applicationModel = getApplicationModel((String)value);
			PlatformImpl.getPlatform().log(Logger.INFO,MasterApplicationInstanceImpl.class,"Application model for application with ID = "+applicationID+" is set to "+applicationModelToString(applicationModel));
		}else{
			applicationModel = DEFAULT_APPLICATION_MODEL; // default model
			PlatformImpl.getPlatform().log(Logger.INFO,MasterApplicationInstanceImpl.class,"No application model specified for application with ID = "+applicationID+", model set to default value "+applicationModelToString(DEFAULT_APPLICATION_MODEL));
		}
		// start the timeout thread
		timeoutThread.start();
	}

	
	/**
	 * Get application model type.
	 * 
	 * @return MasterApplicationID 
	 */
	
	public static byte getApplicationModel(String model){
		if (model.equals(AUTO_APPLICATION_MODEL)) return ApplicationInstanceImpl.AUTO_APPLICATION_MODEL;
		if (model.equals(MANUAL_APPLICATION_MODEL)) return ApplicationInstanceImpl.MANUAL_APPLICATION_MODEL;
		if (model.equals(MAPCYCLE_APPLICATION_MODEL)) return ApplicationInstanceImpl.MAPCYCLE_APPLICATION_MODEL;
		if (model.equals(PERMANENT_APPLICATION_MODEL)) return ApplicationInstanceImpl.PERMANENT_APPLICATION_MODEL;
		else return DEFAULT_APPLICATION_MODEL;
	}
	
	
	/**
	 * Get application model type.
	 * 
	 * @return MasterApplicationID 
	 */
	public static String applicationModelToString(byte model){
		if (model == ApplicationInstanceImpl.AUTO_APPLICATION_MODEL) return MasterApplicationInstanceImpl.AUTO_APPLICATION_MODEL;
		if (model == ApplicationInstanceImpl.MANUAL_APPLICATION_MODEL) return MasterApplicationInstanceImpl.MANUAL_APPLICATION_MODEL;
		if (model == ApplicationInstanceImpl.MAPCYCLE_APPLICATION_MODEL) return MasterApplicationInstanceImpl.MAPCYCLE_APPLICATION_MODEL;
		if (model == ApplicationInstanceImpl.PERMANENT_APPLICATION_MODEL) return MasterApplicationInstanceImpl.PERMANENT_APPLICATION_MODEL;
		return "unknown";
	}
	
	
    /**
	 * Returns the ID of the Master Application Instance.
	 * 
	 * @return MasterApplicationID 
	 */
	public int getMasterApplicationID() {
		return masterApplicationID;
	}

	
	/**
	 * Returns the ID of the application managed by the Master Application Instance.
	 * 
	 * @return MasterApplicationID 
	 */
	public int getApplicationID() {
		return applicationID;
	}
	
	
	/**
	 * Returns the ID of the application managed by the Master Application Instance.
	 * 
	 * @return applicationModel 
	 */
	public byte getApplicationModel() {
		return applicationModel;
	}
	

	/**
	 * Returns the timeout of the application managed by the Master Application Instance.
	 * 
	 * @return applicationTimeout 
	 */
	public int getApplicationTimeout() {
		return applicationTimeout;
	}
	
	
	/***************************************************************************************************/
	/******************************  Application Instances Management **************************************/
	/***************************************************************************************************/
	
	/**
	 * Return an unmodifiable view of the application instances list. 
	 * 
	 * @return
	 */
	public List<ApplicationInstance> getApplicationInstancesList(){
		return Collections.unmodifiableList(applicationInstancesList);
	}
	
	/**
	 * Create an Application Instance.
	 * 
	 * @param ownerSession the Session object representing the actor who create the Application Instance.
	 * @param minActors represents the minimum number of actors required to start the Application Instance.
	 * @param maxActors represents the maximum number of actors can play this Application Instance.
	 * @param actors if the Application Instance is private this vector contains the required actors
	 * else if it is a public Application Instance this vector is empty.
	 * 
	 * @return ActorSessionID of the owner of the ApplicationInstance created.
	 */	
	public short createApplicationInstance(Session ownerSession,int minActors, int maxActors, String[] actors) {
	    ApplicationInstance appInstance = 	ApplicationInstanceImpl.createApplicationInstanceImpl(
											this,
											ownerSession,
											minActors,
											maxActors,
											actors,
											applicationPath,
											applicationModel
											);
	    if(appInstance != null){
	    	 // lock on applicationInstancesList
		    synchronized (applicationInstancesList) {
		    	applicationInstancesList.add(appInstance);
			}
			return appInstance.getOwnerActorSessionID();
	    }else{
			PlatformImpl.getPlatform().log(Logger.WARN,MasterApplicationInstanceImpl.class,"No GameLogic can be linked to this ApplicationInstance, make sure you have properly created it and inserted the matching jar file into the \"jar\" folder of ugasp-maven");	
	    	return ApplicationInstanceImpl.INVALID_APPLICATION_INSTANCE;
	    }
	   
	}
	
	
	/**
	 * Returns the required ApplicationInstance object.
	 * 
	 * @param applicationInstanceID the ID of the required Application Instance
	 * @return 	applicationInstance the object representing the Application Instance
	 * 			null if there is not any application instance with applicationInstanceID
	 */
	public ApplicationInstance getApplicationInstance(int applicationInstanceID) {	
		// lock on appicationInstancesList
	    synchronized (applicationInstancesList) {
	    	for(ApplicationInstance applicationInstance : applicationInstancesList){
	    		if (applicationInstance.getApplicationInstanceID() == applicationInstanceID){
	    			return applicationInstance;
	    		}
	    	}
	    }
		return null;
	}

	
	/**
	 * Determine if the Application Instance associated with applicationInstanceID is hold by this
	 * Master Application.
	 * 
	 * @param applicationInstanceID
	 * 
	 * @return boolean
	 */
	public boolean containsApplicationInstance(int applicationInstanceID) {
		// lock on appicationInstancesList
	    synchronized (applicationInstancesList) {
	    	for (ApplicationInstance applicationInstance : applicationInstancesList) {
				if(applicationInstance.getApplicationInstanceID() == applicationInstanceID){
					return true;
				}
			}
	    }
		return false;
	}
	
	
	/**
	 * Remove the required Application Instance.
	 * 
	 * @param applicationInstanceID
	 * 
	 * @return boolean telling us if the remove was correctly done or not.
	 */
	public boolean removeApplicationInstance(int applicationInstanceID) {
		// lock on appicationInstancesList
	    synchronized (applicationInstancesList) {
	    	for (int i = 0; i < applicationInstancesList.size(); i++) {
				if(applicationInstancesList.get(i).getApplicationInstanceID() == applicationInstanceID){
					// destroy application instance
					applicationInstancesList.get(i).destroyServerInstance();
					// release applicationInstanceID
					PlatformImpl.getPlatform().releaseAIID(applicationInstancesList.get(i).getApplicationInstanceID());
					// remove instance from the list
					applicationInstancesList.remove(i);
					return true;
				}
			}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"Cannot remove ApplicationInstance with ID = "+applicationInstanceID+" since it cannot be retrieved");
		return false;
	}
	

	/**
	 * Returns the ApplicationInstance object containing the ActorSession linked to the Session.
	 * 
	 * @param aSID the session ID
	 * @return the ApplicationInstance object
	 */
	public ApplicationInstance getApplicationInstanceFromActorSessionID(short actorSessionID) {
		// lock on applicationInstancesList
	    synchronized (applicationInstancesList) {
	    	for(ApplicationInstance applicationInstance : applicationInstancesList){
	    		if (applicationInstance.containsActorSession(actorSessionID)){
					return applicationInstance;
				}
	    	}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"There isn't any ApplicationInstance matching with ActorSession with ID = "+actorSessionID);
		return null;
	}
	
	
	/**
	 * Returns the ID of the first currently joinable ApplicationInstance.
	 * 
	 * @return 	the ID of an ApplicationInstance object
	 * 			0 if there is not any joinable application instance
	 */
	public int getJoinableApplicationInstanceID() {
		// lock on appicationInstancesList
	    synchronized (applicationInstancesList) {
	    	for(ApplicationInstance applicationInstance : applicationInstancesList){
	    		if (applicationInstance.isJoinable()){
	    			return applicationInstance.getApplicationInstanceID();
	    		}
	    	}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"There isn't any joinable ApplicationInstance");
		return 0;
	}
	
	/***************************************************************************************************/
	/******************************  Sessions Management ***********************************************/
	/***************************************************************************************************/
	
	/**
	 * Return an unmodifiable view of the session list. 
	 * 
	 * @return 
	 */
	public List<Session> getSessionsList(){
		return Collections.unmodifiableList(sessionsList);
	}
	
	
	/**
	 * Create a new session associated with the actor ID, add it to the list of sessions. 
	 * 
	 * @param actorID the actor ID
	 * 
	 * @return the sessionID of the created session
	 */
	public int createNewSession(int actorID) {	
		Session session = new SessionImpl(actorID, System.currentTimeMillis());
		// lock on the list of sessions
		synchronized (sessionsList) {
			sessionsList.add(session);
		}
		return session.getSessionID();
	}
	
	
	/**
	 * Returns the corresponding session with a refreshed timeout
	 * 
	 * @param sessionID the session ID
	 * 
	 * @return the Session object
	 */
	public Session getSession(int sessionID) {	
		// lock on sessionList
	    synchronized (sessionsList) {
	    	for(Session session : sessionsList){
	    		if (session.getSessionID() == sessionID){
					session.refreshTimeout();
					return session;
				}
	    	}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"There isn't any Session matching with ID = "+sessionID);
		return null;
	}
	
	
	/**
	 * Remove the corresponding Session object.
	 * 
	 * @param sID
	 * 
	 * @return boolean
	 */
	public boolean removeSession(int sessionID){
		// lock on appicationInstancesList
	    synchronized (sessionsList) {
	    	for (int i = 0; i < sessionsList.size(); i++) {
				if(sessionsList.get(i).getSessionID() == sessionID){
					// release sessionID
					PlatformImpl.getPlatform().releaseSID(sessionsList.get(i).getSessionID());
					// remove session from the list
					sessionsList.remove(i);
					return true;
				}
			}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"Session with ID = "+sessionID+" cannot be removed since it cannot be retrieved");
		return false;
	}
	
	
	/**
	 * Return the session associated to the actorID.
	 * 
	 * @param aID
	 * @return the Session object associated
	 */
	public Session getSessionFromActorID(int actorID) {
		// lock on sessionList
	    synchronized (sessionsList) {
	    	for(Session session : sessionsList){
	    		if (session.getActorID() == actorID){
					return session;
				}
	    	}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"There isn't any Session matching with Actor with ID = "+actorID);
		return null;
	}
	
	
	/**
	 * Return the session associated to the Actor Session ID with a refreshed timeout.
	 * 
	 * @param actorSessionID the ActorSession ID
	 * @return the Session object associated
	 */
	public Session getSessionFromActorSessionID(short actorSessionID) {
		// lock on sessionList
	    synchronized (sessionsList) {
	    	for(Session session : sessionsList){
	    		if (session.getActorSessionID() == actorSessionID){
					session.refreshTimeout();
					return session;
				}
	    	}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"There isn't any Session matching with ActorSession with ID = "+actorSessionID);
		return null;
	}
	
	
	/**
	 * Unlog the user represented by this ActorSession ID , delete Actor, Session, ActorSession objects
	 * linked to this user, and also the ApplicationInstance would created by this user.
	 * 
	 * @param sID the Session ID
	 */
	public void unlog(int sessionID) {
		Session session = getSession(sessionID);
		if(session != null){
			short actorSessionID = session.getActorSessionID();
			if (actorSessionID!=0){
				// Get the application instance of the user identified by its actorSessionID 
				ApplicationInstance appInstance = getApplicationInstanceFromActorSessionID(actorSessionID);
				// create quit event 
				QuitEventImpl qe = new QuitEventImpl(actorSessionID);
				//  send it to the application instance of the player
				appInstance.onQuitEvent(qe);
			}
			//remove session from session's list
			removeSession(sessionID);
			//remove actor from actor's list
			removeActor(session.getActorID());
		}else{
		    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"There isn't any Session matching with ID = "+sessionID);
		}
	}
	
	
	/**
	 * Determine if the MasterApplicationInstance is the owner of the session associated with the SessionID.
	 * 
	 * @param sessionID the Session ID
	 * 
	 * @return 	true is it is the owner
	 * 			false otherwise
	 */
	public boolean isSessionOwner(int sessionID) {
		if (this.getSession(sessionID) != null){
			return true;
		}else{
			return false;
		}
	}
	
	
	/***************************************************************************************************/
	/******************************  Actors Management ***********************************************/
	/***************************************************************************************************/
	
	
	/**
	 * Return an unmodifiable view of the actors' list. 
	 * 
	 */
	public List<Actor> getActorsList(){
		return Collections.unmodifiableList(actorsList);
	}
	
	
	/**
	 * Add the Actor object actor to the vector of actors in the Master Application.
	 * 
	 * @param actor the Actor object to add
	 */
	public void addActor(Actor actor) {	
		// lock on actors list
		synchronized (actorsList) {
			actorsList.add(actor);
		}
	}
	
	
	/**
	 * Return the Actor object associated with the actorID.
	 * 
	 * @param aID the actor ID 
	 * 
	 * @return the Actor object
	 */
	public Actor getActor(int actorID) {	
		// lock on sessionList
	    synchronized (actorsList) {
	    	for(Actor actor : actorsList){
	    		if (actor.getActorID() == actorID){
	    			return actor;
	    		}
	    	}
	    }
	    PlatformImpl.getPlatform().log(Logger.DEBUG, this.getClass(),"There isn't any Actor matching with ID = "+actorID);
		return null;
	}
	
	
	/**
	 * Remove the Actor object associated with the actorID.
	 * 
	 * @param aID the actor ID
	 * @return boolean representing the success of the operation
	 */
	public boolean removeActor(int actorID){
		// lock on appicationInstancesList
	    synchronized (actorsList) {
	    	for (int i = 0; i < actorsList.size(); i++) {
				if(sessionsList.get(i).getActorID() == actorID){
					// remove actor from the list
					actorsList.remove(i);
					return true;
				}
			}
	    }
		PlatformImpl.getPlatform().log(Logger.DEBUG,MasterApplicationInstanceImpl.class,"Actor with ID = "+actorID+" cannot be removed since it cannot be retrieved");
		return false;
	}

	
	/**
	 * Determine if the Master Application is the owner of the Actor object associated with the actorID.
	 * 
	 * @param aID the actor ID
	 * 
	 * @return boolean 
	 */
	public boolean isActorOwner(int aID) {
		if (this.getActor(aID) != null) return true;
		return false;
	}
	
	
	/***************************************************************************************************/
	/******************************  Runnable routine  *************************************************/
	/***************************************************************************************************/
	
	/*****************************************/
	/********** TIME OUT THREADING ***********/
	/*****************************************/

	
    public void run() {
        while(true){
        	if (sessionsList.size() > 0){
        		PlatformImpl.getPlatform().log(Logger.TRACE,MasterApplicationInstanceImpl.class,"Check sessions for timeouts");
        		checkTimeouts();
        		deleteEmptyApplicationInstances();
        	}
        	try {
        		Thread.sleep(PLAYERTIMEOUT_CHECKING_DELTATIME);
	       	} catch (Throwable t) {
	       		PlatformImpl.getPlatform().log(Logger.ERROR,MasterApplicationInstanceImpl.class,"Timeout threading error ",t);
	       	}
        }
    }

    
    private void checkTimeouts() {
    	// lock on sessions list
    	synchronized (sessionsList) {
    		// check player timeout
        	for (int i = 0; i < sessionsList.size(); i++){
        		Session session = sessionsList.get(i);
        		// player connected to an ApplicationInstance --> application timeout
        		//FIXME What means this test ? When a since should have an actorSessionID == 0 ?
            	if (session.getActorSessionID() > 0){
            		if ((System.currentTimeMillis()- session.getTimeout()) > applicationTimeout){  
            			ApplicationInstance applicationInstance = getApplicationInstanceFromActorSessionID(session.getActorSessionID());
            			if (applicationInstance != null){
            				applicationInstance.onQuitEvent(new QuitEventImpl(session.getActorSessionID()));
    						PlatformImpl.getPlatform().log(Logger.INFO,MasterApplicationInstanceImpl.class,"ActorSession with ID = "+session.getActorSessionID()+" timeouted: Actor Session quit");
            			}
            		}
            	}
            	// player connected to the lobby room --> platform timeout
            	if ((System.currentTimeMillis()- session.getTimeout()) > platformTimeout){
            		// remove actor
            		removeActor(session.getActorID());
            		// remove session
            		sessionsList.remove(i);
            		PlatformImpl.getPlatform().log(Logger.INFO,MasterApplicationInstanceImpl.class,"Session with ID = "+session.getSessionID()+" timeouted: Session removed");
            	}	
    		}
		}
    }
    
    
    /*
     * Check whether an application instance has clients connected to it.
     * If application model is set to "permanent", nothing is done.
     * Otherwise, application instance is removed from the list of application instances. 
     */
    private void deleteEmptyApplicationInstances(){
    	// lock on application instances list
    	synchronized (applicationInstancesList) {
    		// detect then remove empty application instances
        	for (int i = 0; i < applicationInstancesList.size(); i++) {
        		ApplicationInstance applicationInstance = applicationInstancesList.get(i);
        		// check if actor sessions list is empty, and if the application model isn't set to PERMANENT_APPLICATION_MODEL 
        		if (applicationInstance.getActorSessionsList().isEmpty() && (applicationModel != getApplicationModel(PERMANENT_APPLICATION_MODEL))){
        			applicationInstancesList.remove(applicationInstance);
            		PlatformImpl.getPlatform().log(Logger.INFO,MasterApplicationInstanceImpl.class,"Application Instance with ID = "+i+" is deleted since no more client is present in it.");
            	}
    		}
		}
    }
    
    
    /***************************************************************************************************/
	/******************************  Methods inherited from Object *************************************/
	/***************************************************************************************************/
    
    public boolean equals(Object o){
    	if(o == this){
    		return true;
    	}
    	if(!(o instanceof MasterApplicationInstanceImpl)){
    		return false;
    	}
    	MasterApplicationInstanceImpl masterApplicationInstance = (MasterApplicationInstanceImpl)o;
    	return 	masterApplicationInstance.masterApplicationID == masterApplicationID &&
		    	masterApplicationInstance.applicationID == applicationID &&
		    	masterApplicationInstance.applicationPath.equals(applicationPath) &&
		    	masterApplicationInstance.applicationModel == applicationModel &&
		    	masterApplicationInstance.applicationTimeout == applicationTimeout;
    }
    
    
    public int hashCode(){
    	// Values 37 & 17 are from "The effective Java" by Joshua Bloch
    	int result = 17;
    	result = 37 * result + masterApplicationID;
    	result = 37 * result + applicationID;
    	result = 37 * result + applicationPath.hashCode();
    	result = 37 * result + applicationModel;
    	result = 37 * result + applicationTimeout;
    	return result;
    }
  
}
