/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.impl.core;


import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;
import org.apache.felix.ipojo.InstanceManager;

import org.ugasp.core.Actor;
import org.ugasp.core.ApplicationInstance;
import org.ugasp.core.MasterApplicationInstance;
import org.ugasp.core.Platform;
import org.ugasp.core.Session;
import org.ugasp.gamelogic.GameLogicsFactory;
import org.ugasp.logging.Logger;
import org.ugasp.service.internal.system.db.DBManager;
import org.ugasp.service.internal.system.id.IDManager;

/**
 * Platform is the base of the GAming Services Platform(GASP), the container of all GP
 * representing objects. So it contains all MasterApplicationInstance objects in
 * a vector. It hold all of the service instances (Lobby,...) and system
 * instances (DB manager, ID manager, ...).
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project  
 */


public class PlatformImpl implements Platform {

	//iPojo injection fields
	private DBManager dbManager;
	private IDManager idManager;
	private GameLogicsFactory gameLogicsFactory;
	private Logger logger;

	// Other fields
	private List<MasterApplicationInstance> masterApplicationInstances;
	private static PlatformImpl singleton;

	public boolean gaspDBOK = true;
	protected int platformTimeout;
	protected String logFile;
	protected String logLevel;

	private PlatformImpl(){
		log(Logger.INFO,PlatformImpl.class,"uGASP Core instantiated, platform timeout set to: "+platformTimeout+" ms");
		singleton = this;
		// concurrent list of master application instances
		masterApplicationInstances = new Vector<MasterApplicationInstance>();
	}

	public static PlatformImpl getPlatform(){
		return singleton;
	}

	/**
	 * Returns the MasterApplicationInstance associated with the application ID or if it is not
	 * already instantiates, initialize it.
	 * 
	 * @param applicationID the application ID
	 * @return the MasterApplicationInstance object
	 */
	public synchronized MasterApplicationInstance getMasterApplicationInstance(int applicationID) {

		if (!this.containsMasterApplicationInstance(applicationID)){
			MasterApplicationInstance masterApp = new MasterApplicationInstanceImpl(applicationID,platformTimeout);
			this.addMasterApplicationInstance(masterApp);
			return masterApp;
		}
		else{
			MasterApplicationInstance masterApplicationInstance= null;
			for (Iterator<MasterApplicationInstance> iter = enumerateMasterApplicationInstance(); iter.hasNext();){
				masterApplicationInstance = (MasterApplicationInstance) iter.next();
				if (masterApplicationInstance.getApplicationID() == applicationID)
					return masterApplicationInstance;
			}
		}
		log(Logger.WARN,PlatformImpl.class,"No Master Application found for application with ID = "+applicationID);	
		return null;
	}


	/**
	 * Add the created MasterApplication to the MasterApplications vector.
	 * 
	 * @param masterAppInstance
	 * @return boolean
	 */
	private boolean addMasterApplicationInstance(MasterApplicationInstance masterAppInstance) {
		if (this.containsMasterApplicationInstance(masterAppInstance.getApplicationID()))
			return false;
		else{
			// get a lock on the vector of master application instances
			synchronized (masterApplicationInstances) {
				masterApplicationInstances.add(masterAppInstance);
				return true;	
			}
		}
	}


	/**
	 * Remove the MasterApplicationInstance associated with the application ID.
	 * 
	 * @param applicationID the application ID
	 * @return boolean of the operation success
	 */
	public synchronized boolean removeMasterApplicationInstance(int applicationID) {
		MasterApplicationInstance masterApplicationInstance= null;
		for (Iterator<MasterApplicationInstance> iter = enumerateMasterApplicationInstance(); iter.hasNext();){
			masterApplicationInstance = (MasterApplicationInstance) iter.next();
			int appID = masterApplicationInstance.getApplicationID();
			int masterAppID = masterApplicationInstance.getMasterApplicationID();
			if (appID == applicationID){	
				iter.remove();
				this.getIDManager().releaseMAID(masterAppID);
				return true;
			}
		}
		log(Logger.WARN,PlatformImpl.class,"Removal of Master Application cannot be done, as it can't be found for application with ID = "+applicationID);
		return false;
	}


	/**
	 * Determine if the vector of MasterApplicationInstance is empty or not.
	 * 
	 * @return boolean
	 */
	public boolean isMasterApplicationInstanceEmpty() {
		return masterApplicationInstances.isEmpty();
	}


	/**
	 * Return the number of MasterApplicationInstance currently instantiated.
	 * 
	 * @return the number of MasterApplication instances.
	 **/
	public int masterApplicationInstanceSize() {
		return masterApplicationInstances.size();
	}


	/**
	 * Determine if the MasterApplicationInstance of the application associated with the
	 * application ID is initialized or not.
	 * 
	 * @param applicationID
	 * @return boolean
	 */
	public boolean containsMasterApplicationInstance(int applicationID) {
		MasterApplicationInstance masterApplicationInstance= null;
		// get a lock on the vector of master application instances
		synchronized (masterApplicationInstances) {
			for (Iterator<MasterApplicationInstance> iter = enumerateMasterApplicationInstance(); iter.hasNext();){
				masterApplicationInstance = (MasterApplicationInstance) iter.next();
				if (masterApplicationInstance.getApplicationID() == applicationID)
					return true;
			}	
		}
		return false;
	}

	/**
	 * Return the platform ID manager .
	 * 
	 * @return the IDManager object
	 */
	public IDManager getIDManager(){
		return idManager;
	}


	/**
	 * Return the platform DB manager .
	 * 
	 * @return the DBManager object
	 */
	public DBManager getDBManager() {
		return dbManager;
	}


	/**
	 * Return the MasterApplicationInstance containing the sID.
	 * 
	 * @param sID the session ID
	 * @return the MasterApplicationInstance object
	 */
	public MasterApplicationInstance getMasterApplicationInstanceFromSessionID(int sessionID) {
		MasterApplicationInstance masterApplicationInstance= null;
		// get a lock on the vector of master application instances
		synchronized (masterApplicationInstances) {
			for (Iterator<MasterApplicationInstance> iter = enumerateMasterApplicationInstance(); iter.hasNext();){
				masterApplicationInstance = (MasterApplicationInstance) iter.next();
				if (masterApplicationInstance.isSessionOwner(sessionID))
					return masterApplicationInstance;
			}		    	
		}
		// if no master application can be found
		log(Logger.WARN,PlatformImpl.class,"There isn't any MasterApplicationInstance matching with session with ID = "+sessionID);
		return null;
	}


	/**
	 * Return the MasterApplicationInstance containing the aID.
	 * 
	 * @param aID the actor ID
	 * @return the MasterApplicationInstance object
	 */
	public MasterApplicationInstance getMasterApplicationInstanceFromActorID(int actorID){
		MasterApplicationInstance masterApplicationInstance= null;
		if(isMasterApplicationInstanceEmpty()){
			log(Logger.DEBUG,PlatformImpl.class,"There isn't any Master Application Instance on the platform yet");
		}else{
			log(Logger.DEBUG,PlatformImpl.class,"Actor with ID = "+actorID+" is trying to get its matching Master Application Instance among the "+masterApplicationInstanceSize()+" available ones");
			for (Iterator<MasterApplicationInstance> iter = enumerateMasterApplicationInstance(); iter.hasNext();){
				masterApplicationInstance = (MasterApplicationInstance) iter.next();
				if (masterApplicationInstance.isActorOwner(actorID)){
					log(Logger.DEBUG,PlatformImpl.class,"Actor with ID = "+actorID+" is matching with Master Application Instance with ID = "+masterApplicationInstance.getMasterApplicationID());
					break;
				}

			}
		}
		return masterApplicationInstance;
	}


	/**
	 * Log a user in the platform, authenticated by actorID, username and password.
	 * It consists of creating a new Session in the correct MasterApplicationInstance
	 * and return the session ID.
	 * 
	 * @param aID the Actor ID attributed at the first login
	 * @param username
	 * @param password
	 * 
	 * @return the SessionID of the session created or 0 if the authentication informations
	 * are not valid
	 */
	public int login(int actorID, String username, String password) {
		log(Logger.DEBUG,PlatformImpl.class,"Login request from actorID = "+actorID+" with username = "+username);
		int sessionID = alreadyLogged(actorID);
		if (sessionID==0){ // user not already logged --> checking its authentication
			Vector<?> infos = dbManager.getActorIfAuthentificationOK(actorID, username, password);
			Actor actor = new ActorImpl(actorID,((Integer)infos.elementAt(0)).intValue(), (String) infos.elementAt(1), (String) infos.elementAt(2), (String) infos.elementAt(3));
			if (actor != null){ // authentication ok
				log(Logger.DEBUG,PlatformImpl.class,"Login request: creating sessionID for actorID = "+actorID+" with username = "+username);
				MasterApplicationInstance masterApp = this.getMasterApplicationInstance(actor.getApplicationID());
				masterApp.addActor(actor);
				sessionID = masterApp.createNewSession(actorID);
			}
		}else{
			log(Logger.DEBUG,PlatformImpl.class,"ActorID = "+actorID+" with username = "+username+" already logged with sessionID = "+sessionID);
		}
		return sessionID;
	}


	/**
	 * Return the sID of the user requiring to log in if he's already logged, else 0.
	 * 
	 * @param aID the actor ID of the user
	 * @return the session ID
	 */
	private int alreadyLogged(int actorID) {
		MasterApplicationInstance masterApp = this.getMasterApplicationInstanceFromActorID(actorID);
		if (masterApp != null){ // the Actor is already instantiated then an Session too
			log(Logger.TRACE,PlatformImpl.class,"Actor with ID = "+actorID+" is already logged");
			Session session = masterApp.getSessionFromActorID(actorID);
			if(session != null){
				return session.getSessionID();
			}else{
				log(Logger.TRACE,PlatformImpl.class,"There is no session mathing with actorID = "+actorID);
			}
		}else{
			log(Logger.TRACE,PlatformImpl.class,"Actor with ID = "+actorID+" is not logged yet");
		}
		return 0;
	}


	/**
	 * Provide to a user with a valid session ID to join a specific ApplicationInstance of
	 * the application open by the user. 
	 * Then create an ActorSession in the ApplicationInstance and return the ActorSessionID.
	 * 
	 * @param sID the Session ID
	 * @param aIID the ApplicationInstance ID
	 * @return the ActorSession ID
	 */
	public short joinApplicationInstance(int sessionID, int applicationInstanceID) {
		log(Logger.INFO,PlatformImpl.class,"Join Application Instance request from sessionID = "+ sessionID+" to join Application Instance with ID = "+applicationInstanceID);
		short actorSessionID = 0;
		Session session;
		MasterApplicationInstance masterApp = this.getMasterApplicationInstanceFromSessionID(sessionID);
		if (masterApp != null){
			session = masterApp.getSession(sessionID);
			session.refreshTimeout();

			if (session.getActorSessionID() == 0){ // there isn't any an actorSession linked to that session
				ApplicationInstance applicationInstance = masterApp.getApplicationInstance(applicationInstanceID);
				if (applicationInstance != null){
					if (applicationInstance.isJoinable()){
						actorSessionID = applicationInstance.createNewActorSession(
								applicationInstance.getApplicationInstanceID(),
								session);
					}else{
						log(Logger.DEBUG,PlatformImpl.class,"Session with ID = "+sessionID+" could not join ApplicationInstance with ID = "+applicationInstanceID+" since it isn't joinable");
					}
				}else{
					log(Logger.DEBUG,PlatformImpl.class,"There isn't any ApplicationInstance mathing with ID = "+applicationInstanceID);
				}
			}else{
				return rejoinApplicationInstance(masterApp,session.getActorSessionID());
			}
		}else{
			log(Logger.DEBUG,PlatformImpl.class,"There isn't any MasterApplicationInstance linked with session with ID = "+sessionID+", please make sure loggin has been correctly done");
		}
		return actorSessionID;	
	}

	/**
	 * Rejoin the linked Application Instance after a disconnection.
	 * 
	 * @param masterApp
	 * @param actorSessionID
	 * 
	 * @return the actorSessionID 
	 */
	private short rejoinApplicationInstance(MasterApplicationInstance masterApp, short actorSessionID) {
		masterApp.getApplicationInstanceFromActorSessionID(actorSessionID).sendPlayersInfos(actorSessionID);
		return actorSessionID;
	}


	public Iterator<MasterApplicationInstance> enumerateMasterApplicationInstance() {
		return masterApplicationInstances.iterator();
	}


	/**
	 * Provide to a user with a valid session ID to join a random ApplicationInstance
	 * corresponded to the application opened by the user.
	 * 
	 * @param sID the Session ID
	 * @return the ActorSession ID
	 */
	public int joinRandomApplicationInstance(int sessionID) {
		MasterApplicationInstance masterApp = this.getMasterApplicationInstanceFromSessionID(sessionID);
		if(masterApp != null){
			int applicationInstanceID = masterApp.getJoinableApplicationInstanceID();
			return joinApplicationInstance(sessionID,applicationInstanceID);
		}else{
			log(Logger.DEBUG,PlatformImpl.class,"There isn't any MasterApplicationInstance linked with session with ID = "+sessionID+", please make sure loggin has been correctly done");
			return 0;
		}
	}


	/**
	 * Provide to a user with a valid session ID to create an ApplicationInstance with a minimum
	 * start number of actors and a maximum number of actors.
	 * The ApplicationInstance status is:
	 * "public": if the table of actors is empty.
	 * "private": if the table of actors contains some Actor ID.
	 * Returns the ID of the ApplicationInstance created.
	 * 
	 * @param sID the Session ID
	 * @param minActors minimum number of actors for starting
	 * @param maxActors maximum number of actors
	 * @param actors a table of Actor ID
	 * @return ActorSessionID of the owner of the ApplicationInstance created.
	 */
	public short createApplicationInstance(int sessionID, int minActors, int maxActors, String[] actors) {
			log(Logger.INFO,PlatformImpl.class,"Create Application Instance request from sessionID = "+ sessionID+"with at least "+minActors+" actors, and at most "+maxActors+" actors");
			short actorSessionID = 0;
			MasterApplicationInstance masterApp = this.getMasterApplicationInstanceFromSessionID(sessionID);
			if(masterApp != null){
				Session session = masterApp.getSession(sessionID);
				if (session.getActorSessionID() == 0){ 
					log(Logger.DEBUG,PlatformImpl.class,"Session with ID = "+ sessionID+" is not currently linked with an Application Instance, new Application Instance creation");
					actorSessionID =  masterApp.createApplicationInstance(session,minActors,maxActors,actors);
				}else{ 
					log(Logger.DEBUG,PlatformImpl.class,"Session with ID = "+ sessionID+" is already linked with an Application Instance, Application Instance is joined back");
					return rejoinApplicationInstance(masterApp, session.getActorSessionID());
				}
			}else{
				log(Logger.DEBUG,PlatformImpl.class,"There isn't any MasterApplicationInstance linked with session with ID = "+sessionID+", please make sure loggin has been correctly done");
			}
			return actorSessionID;
	}


	/**
	 * This method is called when a user quit the application.
	 * ActorSession, Session, Actor instances corresponding to the user are removed.
	 * 
	 * @param sID the Session ID
	 */
	public void quit(int sessionID){
		log(Logger.INFO,PlatformImpl.class,"Quit platform request from sessionID = "+ sessionID);
		MasterApplicationInstance masterApp = this.getMasterApplicationInstanceFromSessionID(sessionID);
		if(masterApp != null){
			masterApp.unlog(sessionID);
		}else{
			log(Logger.DEBUG,PlatformImpl.class,"There isn't any MasterApplicationInstance linked with session with ID = "+sessionID+", please make sure loggin has been correctly done");
		}
	}


	/**
	 * Return the MasterApplication hosts the Session linked to the ActorSession ID.
	 * 
	 * @param aSID the ActorSessionID
	 * @return the MasterApplication
	 */
	private MasterApplicationInstance getMasterApplicationInstanceFromActorSessionID(short actorSessionID) {
		MasterApplicationInstance masterApplicationInstance = null;
		// get a lock on the vector of master application instance
		synchronized (masterApplicationInstances) {
			for (Iterator<MasterApplicationInstance> iter = enumerateMasterApplicationInstance(); iter.hasNext();){
				masterApplicationInstance = (MasterApplicationInstance) iter.next();
				if (masterApplicationInstance.getApplicationInstanceFromActorSessionID(actorSessionID)!= null){
					return masterApplicationInstance;
				}
			}	
		}
		return null;
	}


	/**
	 * Return the ApplicationInstance containing the actor session ID.
	 * 
	 * @param aSID the actor session ID
	 * @return the ApplicationInstance object
	 */
	public ApplicationInstance getApplicationInstanceFromActorSessionID(short actorSessionID) {
		ApplicationInstance appInstance = null;
		MasterApplicationInstance masterApplication = this.getMasterApplicationInstanceFromActorSessionID(actorSessionID);
		if(masterApplication != null){
			masterApplication.getSessionFromActorSessionID(actorSessionID).refreshTimeout();
			appInstance =  masterApplication.getApplicationInstanceFromActorSessionID(actorSessionID);		
		}else{
			log(Logger.DEBUG,PlatformImpl.class,"Not any Master Application Instance is matching with actorSessionID = "+actorSessionID);
		}
		return appInstance;
	}


	/**
	 * Return the platform timeout
	 */
	public int getPlatformTimeout() {
		return platformTimeout;
	}

	public void log(byte logLevel, Class<?> clazz, String message){
		logger.log(logLevel,clazz, message);
	}


	public void log(byte logLevel, Class<?> clazz, String message, Throwable t){
		logger.log(logLevel,clazz, message, t);
	}

	/*****************************************************************/
	/** Methods to route ID Manager calls from other bundle classes **/
	/*****************************************************************/

	protected int requestMAID(){return idManager.generateMAID();}
	protected int requestAIID(){return idManager.generateAIID();}
	protected int requestSID(){return idManager.generateSID();}
	protected short requestASID(){return idManager.generateASID();}

	protected void releaseMAID(int mAID){idManager.releaseMAID(mAID);}
	protected void releaseAIID(int aIID){idManager.releaseAIID(aIID);}
	protected void releaseSID(int sID){idManager.releaseSID(sID);}
	protected void releaseASID(short aSID){idManager.releaseASID(aSID);}

	/*********************************************************************/
	/** Methods to route FactoryManager calls from other bundle classes **/
	/*********************************************************************/

	protected InstanceManager newGameLogic(int appID){return gameLogicsFactory.newGameLogicInstance(appID);}
	protected Hashtable<?,?> getApplicationProperties(int appID){return gameLogicsFactory.getApplicationProperties(appID);}



}
