package org.ugasp.impl.protocol.rmai;


import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.io.IOException;

import org.ugasp.protocol.rmai.RMAISerializable;
import org.ugasp.protocol.rmai.RMAISerializer;

import org.xmlpull.v1.XmlSerializer;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;


/**
 *  This class is made to serialize and deserialize Object handled by RMAI
 *  in XML format, using tags defined in RMAISerializer interface.
 * 
 */
class RMAISerializerImpl implements RMAISerializer {
	
	
	static SimpleDateFormat dateFormat = new SimpleDateFormat(DATETIME_FORMAT);
	private final static String MISSING_NO_ARGUMENT_CONSTRUCTOR_EXCEPTION_MESSAGE;
	private final static String NOT_ACCESSIBLE_NO_ARGUMENT_CONSTRUCTOR_EXCEPTION_MESSAGE;

	// Initialize exceptions messages
	static{
		MISSING_NO_ARGUMENT_CONSTRUCTOR_EXCEPTION_MESSAGE			= "You MUST add a no-argument constructor to your RMAISerializable object to use it within RMAI.";
		NOT_ACCESSIBLE_NO_ARGUMENT_CONSTRUCTOR_EXCEPTION_MESSAGE 	= "The no-argument constructor of your RMAISerializable object MUST be public.";
	}

	
	public void serialize(XmlSerializer serializer, Object object) throws Exception, IOException {
		// check for scalar types:
		if (object instanceof Byte) {
			serializer.startTag(null, TYPE_I1).text(object.toString()).endTag(null, TYPE_I1);
		} else
		if (object instanceof Short) {
			serializer.startTag(null, TYPE_I2).text(object.toString()).endTag(null, TYPE_I2);
		} else
		if (object instanceof Integer) {
			serializer.startTag(null, TYPE_I4).text(object.toString()).endTag(null, TYPE_I4);
		} else
		if (object instanceof Long) {
			serializer.startTag(null, TYPE_I8).text(object.toString()).endTag(null, TYPE_I8);
		} else
		if (object instanceof Float) {
			serializer.startTag(null, TYPE_FLOAT).text(object.toString()).endTag(null, TYPE_FLOAT);
		} else
		if (object instanceof Double) {
			serializer.startTag(null, TYPE_DOUBLE).text(object.toString()).endTag(null, TYPE_DOUBLE);
		} else
		if (object instanceof Boolean) {
			Boolean bool = (Boolean) object;
			String boolStr = bool.booleanValue() ? "1" : "0";
			serializer.startTag(null, TYPE_BOOLEAN).text(boolStr).endTag(null, TYPE_BOOLEAN);
		} else
		if (object instanceof String) {
			serializer.startTag(null, TYPE_STRING).text(object.toString()).endTag(null, TYPE_STRING);
		} else
		if (object instanceof Date || object instanceof Calendar) {
			String dateStr = dateFormat.format(object);
			serializer.startTag(null, TYPE_DATE_TIME_ISO8601).text(dateStr).endTag(null, TYPE_DATE_TIME_ISO8601);
		} else
		if (object instanceof List) {
			serializer.startTag(null, TYPE_ARRAY).startTag(null, TAG_DATA);
			List<Object> list = (List<Object>) object;
			Iterator<Object> iter = list.iterator();
			while (iter.hasNext()) {
				Object o = iter.next();
				serializer.startTag(null, TAG_VALUE);
				serialize(serializer, o);
				serializer.endTag(null, TAG_VALUE);
			}
			serializer.endTag(null, TAG_DATA).endTag(null, TYPE_ARRAY);
		} else
		if (object instanceof Object[]) {
			serializer.startTag(null, TYPE_ARRAY).startTag(null, TAG_DATA);
			Object[] objects = (Object[]) object;
			for (int i=0; i<objects.length; i++) {
				Object o = objects[i];
				serializer.startTag(null, TAG_VALUE);
				serialize(serializer, o);
				serializer.endTag(null, TAG_VALUE);
			}
			serializer.endTag(null, TAG_DATA).endTag(null, TYPE_ARRAY);
		} else
		if (object instanceof Map) {
			serializer.startTag(null, TYPE_STRUCT);
			Map<String, Object> map = (Map<String, Object>) object;
			Iterator<Entry<String, Object>> iter = map.entrySet().iterator();
			while (iter.hasNext()) {
				Entry<String, Object> entry = iter.next();
				String key = entry.getKey();
				Object value = entry.getValue();

				serializer.startTag(null, TAG_MEMBER);
				serializer.startTag(null, TAG_NAME).text(key).endTag(null, TAG_NAME);
				serializer.startTag(null, TAG_VALUE);
				serialize(serializer, value);
				serializer.endTag(null, TAG_VALUE);
				serializer.endTag(null, TAG_MEMBER);
			}
			serializer.endTag(null, TYPE_STRUCT);
		} else
		if (object instanceof RMAISerializable) {
			RMAISerializable serializable = (RMAISerializable) object;
			// retrieve object as a map
			Map<String, Object> map = serializable.getSerializable();
			// tag for the RMAISerializable
			serializer.startTag(null, TYPE_RMAI_SERIALIZABLE);
			// get the "useful" class name needed to retrieve class on the client side
			String classNameToSerialize = RMAISerializableUtil.getClassNameToStore(object.getClass());
			// tag for the class name of the RMAISerializable instance			
			serializer.startTag(null, TYPE_STRING).text(classNameToSerialize).endTag(null, TYPE_STRING);
			// serializing the map of the RMAISerializable
			serializer.startTag(null, TYPE_STRUCT);
			Iterator<Entry<String, Object>> iter = map.entrySet().iterator();
			while (iter.hasNext()) {
				Entry<String, Object> entry = iter.next();
				String key = entry.getKey();
				Object value = entry.getValue();

				serializer.startTag(null, TAG_MEMBER);
				serializer.startTag(null, TAG_NAME).text(key).endTag(null, TAG_NAME);
				serializer.startTag(null, TAG_VALUE);
				serialize(serializer, value);
				serializer.endTag(null, TAG_VALUE);
				serializer.endTag(null, TAG_MEMBER);
			}
			// end tag for map section
			serializer.endTag(null, TYPE_STRUCT);
			// end tag for the whole serializable section
			serializer.endTag(null, TYPE_RMAI_SERIALIZABLE);
		} else {
			throw new IOException("Cannot serialize " + object);
		}
	}
	
	
	/**
	 * @throws	ClassNotFoundException when trying to deserialize an object which class that could be found
	 * @throws	InstantiationException when trying to deserialize an object which has no nullary constructor 	
	 * @throws	IllegalAccessException when trying to deserialize an object which nullary constructor is not public
	 */
	public Object deserialize(XmlPullParser parser) throws 	XmlPullParserException, IOException,
															ClassNotFoundException, IllegalAccessException,
															InstantiationException{
		parser.require(XmlPullParser.START_TAG, null, TAG_VALUE);

		if (parser.isEmptyElementTag()) {
			// degenerated <value />, return empty string
			return "";
		}
		
		Object obj = null;
		boolean hasType = true;
		String typeNodeName = null;
		try {
			parser.nextTag();
			typeNodeName = parser.getName();
			if (typeNodeName.equals(TAG_VALUE) && parser.getEventType() == XmlPullParser.END_TAG) {
				// empty <value></value>, return empty string
				return "";
			}
		} catch (XmlPullParserException e) {
			hasType = false;
		}
		if (hasType) {
			if (typeNodeName.equals(TYPE_I1)) {
				String value = parser.nextText();
				obj = Byte.parseByte(value);
			} else
			if (typeNodeName.equals(TYPE_I2)) {
				String value = parser.nextText();
				obj = Short.parseShort(value);
			} else
			if (typeNodeName.equals(TYPE_I4)) {
				String value = parser.nextText();
				obj = Integer.parseInt(value);
			} else
			if (typeNodeName.equals(TYPE_I8)) {
				String value = parser.nextText();
				obj = Long.parseLong(value);
			} else
			if (typeNodeName.equals(TYPE_FLOAT)) {
				String value = parser.nextText();
				obj = Float.parseFloat(value);
			} else
			if (typeNodeName.equals(TYPE_DOUBLE)) {
				String value = parser.nextText();
				obj = Double.parseDouble(value);
			} else
			if (typeNodeName.equals(TYPE_BOOLEAN)) {
				String value = parser.nextText();
				obj = value.equals("1") ? Boolean.TRUE : Boolean.FALSE;
			} else
			if (typeNodeName.equals(TYPE_STRING)) {
				obj = parser.nextText();
			} else
			if (typeNodeName.equals(TYPE_DATE_TIME_ISO8601)) {
				String value = parser.nextText();
				try {
					obj = dateFormat.parseObject(value);
				} catch (ParseException e) {
					throw new IOException("Cannot deserialize dateTime " + value); 
				}
			} else
			if (typeNodeName.equals(TYPE_ARRAY)) {
				parser.nextTag(); // TAG_DATA (<data>)
				parser.require(XmlPullParser.START_TAG, null, TAG_DATA);
	
				parser.nextTag();
				List<Object> list = new ArrayList<Object>();
				while (parser.getName().equals(TAG_VALUE)) {
					list.add(deserialize(parser));
					parser.nextTag();
				}
				parser.require(XmlPullParser.END_TAG, null, TAG_DATA);
				parser.nextTag(); // TAG_ARRAY (</array>)
				parser.require(XmlPullParser.END_TAG, null, TYPE_ARRAY);
				obj = list.toArray();
			} else
			if (typeNodeName.equals(TYPE_STRUCT)) {
				parser.nextTag();
				Map<String, Object> map = new HashMap<String, Object>();
				while (parser.getName().equals(TAG_MEMBER)) {
					String memberName = null;
					Object memberValue = null;
					while (true) {
						parser.nextTag();
						String name = parser.getName();
						if (name.equals(TAG_NAME)) {
							memberName = parser.nextText();
						} else
						if (name.equals(TAG_VALUE)) {
							memberValue = deserialize(parser);
						} else {
							break;
						}
					}
					if (memberName != null && memberValue != null) {
						map.put(memberName, memberValue);
					}
					parser.require(XmlPullParser.END_TAG, null, TAG_MEMBER);
					parser.nextTag();
				}
				parser.require(XmlPullParser.END_TAG, null, TYPE_STRUCT);
				obj = map;
			} else 
			if (typeNodeName.equals(TYPE_RMAI_SERIALIZABLE)){
				// move parser on next tag to read class name (String)
				parser.nextTag();
				String className = parser.nextText();
				// String end tag
				parser.nextTag();
				// retrieve the class name on the platform side
				String platformSideClassName = RMAISerializableUtil.getPlatformSideClassName(className);
				// instantiation of RMAI serializable object
				obj = RMAISerializableUtil.createInstanceOfRMAISerializable(platformSideClassName);
				// Struct tag
				parser.getName();
				parser.nextTag();
				// Member tag
				parser.getName();
				Map<String, Object> map = new HashMap<String, Object>();
				while (parser.getName().equals(TAG_MEMBER)) {
					String memberName = null;
					Object memberValue = null;
					while (true) {
						parser.nextTag();
						String name = parser.getName();
						if (name.equals(TAG_NAME)) {
							memberName = parser.nextText();
						} else
							if (name.equals(TAG_VALUE)) {
								memberValue = deserialize(parser);
							} else {
								break;
							}
					}
					if (memberName != null && memberValue != null) {
						map.put(memberName, memberValue);
					}
					parser.require(XmlPullParser.END_TAG, null, TAG_MEMBER);
					parser.nextTag();
				}
				// End tag TYPE_STRUCT
				parser.require(XmlPullParser.END_TAG, null, TYPE_STRUCT);
				// End tag TYPE_RMAI_SEERIALIZABLE
				parser.nextTag();
				parser.require(XmlPullParser.END_TAG, null, TYPE_RMAI_SERIALIZABLE);
				// initialize object with map
				((RMAISerializable)(obj)).initialize(map);
			} else{
				throw new IOException("Cannot deserialize " + parser.getName());
			}
		} else {
			// TYPE_STRING (<string>) is not required
			obj = parser.getText();
		}
		parser.nextTag(); // TAG_VALUE (</value>)
		parser.require(XmlPullParser.END_TAG, null, TAG_VALUE);
		return obj;
	}
	
	
	/**
	 * Set the folder where RMAISerializable objects are located.
	 * All RMAISerializable objects MUST be located with in the same folder.
	 * 
	 * @param folderName
	 */
	static void setRMAISerializableObjectsFolder(String folderName){
		RMAISerializableUtil.RMAI_SERIALIZABLE_OBJECTS_FOLDER = folderName;
	}

	/**
	 * This class contains methods needed to serialize and deserialize RMAISerializable objects.
	 * 
	 * @author Gabriel Adgeg (gabriel.adgeg@telecom-sudparis.eu)
	 */
	private static class RMAISerializableUtil{
		
		// The folder in which every RMAISerializable objects should be located on the gamelogic side
		private static String RMAI_SERIALIZABLE_OBJECTS_FOLDER;
		private static String RMAI_SERIALIZABLE_OBJECTS_FOLDER_NO_SET;
		private static String RMAI_SERIALIZABLE_OBJECTS_FOLDER_NO_SET_MESSAGE;
		private static String RMAI_SERIALIZABLE_NOT_FOUND_EXCEPTION_MESSAGE;
		
		// Initialize final Strings
		static{
			RMAI_SERIALIZABLE_OBJECTS_FOLDER_NO_SET			=   "Folder path not set.";
			RMAI_SERIALIZABLE_OBJECTS_FOLDER 				= 	RMAI_SERIALIZABLE_OBJECTS_FOLDER_NO_SET;
			RMAI_SERIALIZABLE_OBJECTS_FOLDER_NO_SET_MESSAGE =	"Folder where RMAISerializable Object are located must be specified " +
																"using setRMAISerializableObjectsFolder(String folderName) on your RMAIMgr";
			RMAI_SERIALIZABLE_NOT_FOUND_EXCEPTION_MESSAGE 	= 	"Cannot find the RMAISerializable objects. It could be:\n"+
																"\u2022 either because your class is only located on clients sides and not on the game logic side\n"+
																"\u2022 or because you did not call setRMAISerializableObjectsFolder(String folderName) on your RMAIMgr " +
																"to specify where RMAISerializable objects are located. "+
																"Please refer to RMAISerializable javadoc and its exemple.";
		}
		
		/**
		 * Create an instance of an implementation of RMAISerializable.
		 * 
		 * @param 	rmaiSerializableImplementationName the class name of the implementation of RMAISerializable that should be initialized
		 * @return 	an instance of the class designated by param rmaiSerializableImplementationName
		 * @throws	ClassNotFoundException if no definition for the class with the specified name could be found
		 * @throws	InstantiationException if the class hasn't a no-argument constructor 	
		 * @throws	IllegalAccessException if the no-argument constructor is not public
		 */
		public static RMAISerializable createInstanceOfRMAISerializable(String rmaiSerializableImplementationName) throws ClassNotFoundException, IllegalAccessException, InstantiationException{
			RMAISerializable serializable = null;
			try{
				Class<?> cl = Class.forName(rmaiSerializableImplementationName);
				//return (RMAISerializable)cl.newInstance();
				serializable = (RMAISerializable)cl.newInstance();
			}catch(InstantiationException e){
				throw new InstantiationException(MISSING_NO_ARGUMENT_CONSTRUCTOR_EXCEPTION_MESSAGE);
			}catch(IllegalAccessException e){
				throw new InstantiationException(NOT_ACCESSIBLE_NO_ARGUMENT_CONSTRUCTOR_EXCEPTION_MESSAGE);
			}catch(ClassNotFoundException e){
				throw new ClassNotFoundException(RMAI_SERIALIZABLE_NOT_FOUND_EXCEPTION_MESSAGE);
			}
			return serializable;
		}
		
		
		/**
		 * Get the String to store in order to retrieve the name of the implementation of RMAISerializable
		 * on the platform side.
		 * 
		 * @param 	the class implementing RMAISerializable
		 * @return	the class name stored that shall be stored during serialization process
		 * @throws	Exception if class is not located in the RMAI_SERIALIZABLE_OBJECTS_FOLDER 
		 * 			or if RMAI_SERIALIZABLE_OBJECTS_FOLDER is not set.
		 */
		public static String getClassNameToStore(Class<?>cl)throws Exception{
			String res = null;
			String className = cl.getName();
			String simpleClassName = cl.getSimpleName();
			if(RMAI_SERIALIZABLE_OBJECTS_FOLDER.equals(RMAI_SERIALIZABLE_OBJECTS_FOLDER_NO_SET) || !className.startsWith(RMAI_SERIALIZABLE_OBJECTS_FOLDER)){
				throw new Exception(RMAI_SERIALIZABLE_OBJECTS_FOLDER_NO_SET_MESSAGE);
			}else if(className.equals(RMAI_SERIALIZABLE_OBJECTS_FOLDER+"."+simpleClassName)){
				res = simpleClassName;
			}
			return res;
		}
		
		/**
		 * Get the class name of a RMAISerializable implementation on the platform side from its
		 * name sent by client.
		 * 
		 * @param 	classNameStored the name of the RMAISerializable implementation stored by the client
		 * @return	the name of the same class on the platform side
		 */
		public static String getPlatformSideClassName(String classNameStored){
			return RMAI_SERIALIZABLE_OBJECTS_FOLDER+"."+classNameStored;
		}
		
	}

}
