/***
import java.sql.DriverManager;
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.impl.service.internal.system.db.mysql;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Vector;

import org.osgi.framework.BundleContext;
import org.ugasp.logging.Logger;

public class DBManagerImpl implements org.ugasp.service.internal.system.db.DBManager {
	//iPOJO injection field
	private Logger logger;

	private Connection connection;
	private boolean isDBConnectivityOK = false;

	private static final String DB_DRIVER_NAME 	= "org.gjt.mm.mysql.Driver";
	private static final String DB_URL 			= "jdbc:mysql://localhost:3306/gaspdb";
	private static final String DB_USER 		= "gasp";
	private static final String DB_USER_PWD 	= "gasppwd";
	private static final int 	ALLOWED_USER_ID = 0;

	public DBManagerImpl(BundleContext bc){
		try {
			Driver driver = (Driver) bc.getBundle().loadClass(DB_DRIVER_NAME).newInstance();
			connection = DriverManager.getConnection(DB_URL, DB_USER, DB_USER_PWD);
			isDBConnectivityOK = true;
			logger.log(Logger.INFO, this.getClass(),"Database Manager instantiated");
		} catch(Throwable t){
			logger.log(Logger.ERROR, this.getClass(),"Exception on the loading of DataBase or of the MySQL Driver, please check that your MYSQL Database is properly connected",t);
		}
	}


	public void stop(){
		try{
			isDBConnectivityOK = false;
			if (connection!=null){
				connection.close();
				if(logger != null){
					logger.log(Logger.DEBUG, this.getClass(),"MySQL connection closed");
				}
			}		
		}
		catch (Throwable t){
			if(logger != null){
				logger.log(Logger.ERROR, this.getClass(),"Exception on the closing of MYSQL Database",t);
			}
		}
	}


	/**
	 * Verify if the connection to the GASP DB is ok.
	 * 	
	 * @return boolean
	 */
	public boolean isDBConnectionOK(){ 
		return isDBConnectivityOK;
	}


	/**
	 * Generate a DB persistant Actor ID. 
	 * 
	 * @return the actor ID
	 */ 
	public int generateActorID() {
		int minFreeActorID = 1;
		try {
			// Statement for "default" result set, i.e. with type = RT.TYPE_FORWARD_ONLY 
			// and concurrency = RT.CONCUR_READ_0NLY 
			// --> Anyway, we just want to ensure database coherency, and not Resultset one 
			Statement statement = (Statement) connection.createStatement();

			// search the lowest free actorID in the table actors
			ResultSet result = statement.executeQuery("SELECT actorID FROM actors ORDER BY actorID;");
			while (result.next()){
				int sInt = result.getInt("actorID");
				if (sInt == minFreeActorID) minFreeActorID++;
				if (sInt > minFreeActorID) return Short.parseShort(""+minFreeActorID);
			}
			return minFreeActorID;
		} 
		catch (Throwable t) {
			logger.log(Logger.ERROR, this.getClass(),"Exception on the generation of an actorID",t);
		}
		return 0;
	}


	/**
	 * Verify if the login authentication of an Actor is valid 
	 * then return the authenticated actor or null.
	 * 
	 * @param aID the actor ID
	 * @param username 
	 * @param password
	 * @return the Actor instance object
	 */
	public Vector getActorIfAuthentificationOK(int aID, String username,String password) {
		try{
			Statement statement = (Statement) connection.createStatement();

			// hash the password
			ResultSet result1 = statement.executeQuery("SELECT PASSWORD('"+password+"');");
			result1.first();
			String passwordHashed = result1.getString(1);

			// get the actor n-uple matching the aID
			String query = "SELECT * FROM actors WHERE actorID='"+aID+"';";
			ResultSet result2 = statement.executeQuery(query);

			// test if the login/pass matching
			if (result2.first()){
				String user = result2.getString("username");
				String pwd = result2.getString("password");
				String pseudo = result2.getString("lastPseudo");
				if (username.equals(user)&&passwordHashed.equals(pwd)){

					//Actor actor = new ActorImpl(aID,result2.getInt("appID"),user,pwd,pseudo);
					//return actor;
					Vector v = new Vector();
					v.addElement(result2.getInt("appID"));
					v.addElement(user);
					v.addElement(pwd);
					v.addElement(pseudo);

					return v;
				}
			}
		}catch (Throwable t) {
			logger.log(Logger.ERROR, this.getClass(),"Exception on the getting of properly anthentificated actor",t);
		}
		return null;
	}

	/**
	 * Register a new actor for the double (uid,appID) if the user have the appropriate rights.
	 * 
	 * @param uid
	 * @param appID
	 * @param username
	 * @param password
	 * @return the Actor ID
	 */
	public int registerNewActor(int uid, int appID, String username, String password) {
		int actorID = 0;
		try {
			if (connection == null){
				logger.log(Logger.INFO, this.getClass(), "Database Connection cannot succeed as it has a null value");
			}else{
				logger.log(Logger.INFO, this.getClass(), "Database Connection succeed");
			}
			Statement statement = (Statement) connection.createStatement();
			if (statement == null){
				logger.log(Logger.INFO, this.getClass(), "Null statement on registering of new actor");
			}
			int aIDRegistered = userAlreadyRegistered(statement,uid,appID,username);
			if (aIDRegistered!=0){
				logger.log(Logger.INFO, this.getClass(), "Client already registered with actorID = "+aIDRegistered);
				return aIDRegistered;
			}
			else {
				if (this.isUserAllowed(statement,uid,appID)){
					logger.log(Logger.INFO, this.getClass(), "Null statement on registering of new actor");
					// generate an actorID
					actorID = this.generateActorID();
					// hash the password
					ResultSet result1 = statement.executeQuery("SELECT PASSWORD('"+password+"');");
					if (result1 == null){
						logger.log(Logger.INFO, this.getClass(), "Hash password is null on registering of new actor");
					}
					result1.first();
					String pwdHashed = result1.getString(1);
					// insert the new actor n-uple;
					String query = "INSERT INTO actors VALUES " +
					"("+ actorID +","+ uid +","+ appID +",'"+ username +"','"+ pwdHashed +"','Player');";
					statement.execute(query);
					return actorID;
				}
			}
		} catch (Throwable t) {
			logger.log(Logger.ERROR, this.getClass(),"Exception on the registration of new actor",t);
		}
		return 0;
	}

	/*
	 * Determine by search the table rights if the user have already an actorID.
	 * 
	 * @param statement the Statement of DB
	 * @param uid the User ID
	 * @param appID the Application ID
	 * @param username the user name
	 * @return boolean
	 */
	private int userAlreadyRegistered(Statement statement, int uid, int appID, String username) {
		try{
			String query = "SELECT * FROM actors WHERE userID='"+ uid +"' AND appID='"+ appID +"' AND username='"+ username +"'";
			ResultSet result = statement.executeQuery(query);
			if (result.first()) return result.getInt("actorID");
			else return 0;
		}catch (Throwable t) {
			logger.log(Logger.ERROR, this.getClass(),"Exception on the retrieving of an already registered user",t);
		}
		return 0;
	}


	/*
	 * Determine by search the table rights if the user have the rights
	 * to play the application appID.
	 * If user ID matches with ALLOWED_USER_ID, method always return true.
	 * 
	 * @param statement the Statement of DB
	 * @param uid the User ID
	 * @param appID the Application ID
	 * @return boolean
	 */
	private boolean isUserAllowed(Statement statement, int uid, int appID) {
		if(uid == ALLOWED_USER_ID){
			return true;
		}
		try{
			String query = "SELECT * FROM rights WHERE userID='"+ uid +"' AND appID='"+ appID +"'";
			ResultSet result = statement.executeQuery(query);
			if (result.first()) return true;
		}catch (Throwable t) {
			logger.log(Logger.ERROR, this.getClass(),"Exception on the verification of user's right for an application",t);
		}
		return false;
	}

	/**
	 * Saves in the DB the last used pseudo.
	 * 
	 * @param actorID
	 * @param pseudoName
	 */
	public void saveLastUsedPseudo(int actorID, String pseudoName) {
		try{
			Statement statement = (Statement) connection.createStatement();
			String query = "UPDATE actors SET lastPseudo='"+pseudoName+"' WHERE actorID='"+ actorID +"'";
			statement.execute(query);
		}catch (Throwable t) {
			logger.log(Logger.ERROR, this.getClass(),"Exception on the saving of last pseudo used by an user",t);
		}
	}
}
