/***
 * uGASP: an OSGi middleware dedicated to ubiquitous multiplayer gaming.
 * Copyright (C) 2008 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.ugasp.impl.gamelogic;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

import org.ugasp.eventmodel.event.*;
import org.ugasp.eventmodel.event.impl.MooDSEventImpl;
import org.ugasp.eventmodel.event.impl.RMAIEventImpl;
import org.ugasp.gamelogic.GameLogic;
import org.ugasp.protocol.rmai.InvokedMethod;
import org.ugasp.protocol.rmai.RMAIUtils;
import org.ugasp.protocol.moods.CustomTypes;
import org.ugasp.core.ApplicationInstance;
import org.ugasp.logging.Logger;

/**
 * A game logic running on uGASP must implement this abstract class.
 * 
 * @author PELLERIN Romain (romain.pellerin@ubidreams.com) - GASP Project 
 */
public abstract class GameLogicImpl implements GameLogic{
	
	
	/**
	 * Field to handle MooDS CustomTypes class of extending game logic class.
	 */
	protected CustomTypes customTypes;
    
	/**
	 * Field to link game logic with associated core application instance.
	 */
	protected ApplicationInstance applicationInstance;
    
    
    protected Map<String,String> optionalInstanceInfos = new HashMap<String, String>();
    
 	
    /**
     * Link the application instance to the game logic.
     * 
     * @param applicationInstance
     */
    public void setOwnerAI(ApplicationInstance applicationInstance){
        this.applicationInstance = applicationInstance;
    }
    
    
    /**
     * Retrieve the actor ID linked to the actor session ID.
     * (persistence) Provide the storage of player game informations. 
     * 
     * @param actorSessionID
     * @param dataEvent the DataEvent to send
     */
    public int retrieveActorID(short actorSessionID){
        return applicationInstance.getActorSession(actorSessionID).getActorID();
    }
    
    
    /**
     * Retrieve the application ID.
     * (persistence) Provide the storage of player game informations. 
     */
    public int retrieveApplicationID(){
        return applicationInstance.retrieveApplicationID();
    }
    
    
    /**
	 * Return the pseudo of a user from its actorSessionId.
	 * 
	 * @return	the pseudo of user (Beware, pseudo is NOT the user name).
	 */
    public String getPseudoByActorSessionId(short actorSessionId){
    	String pseudo = null;
    	if(applicationInstance != null){
    		pseudo = applicationInstance.getPseudoByActorSessionId(actorSessionId);
    	}
    	return pseudo;
    }
    
    
    /**
	 * Return actorSessionId of a user from its pseudo.
	 * 
	 * @return -1 if the no actorSessionId could be retrieved for this pseudo			
	 */
    public short getActorSessionIdByPseudo(String pseudo){
    	short actorSessionId = -1;
    	if(applicationInstance != null){
    		actorSessionId = applicationInstance.getActorSessionIdByPseudo(pseudo);
    	}
    	return actorSessionId;
    }
    
    
    /**
     * Get a map of optional in game informations. This informations can be retrieved in ApplicationInstanceInfos object 
     * within the Lobby phase.
     * By example, this map should be used in some game to add data such as mission level matching with one instance.
     * In this case, an entry of map should be :
     * key = "level" - value = "Mission to Mars"
     * 
     * @return 	the Map containing every in game optional information
     */
    public Map<String,String> getOptionalInstanceInfos(){
    	return optionalInstanceInfos;
    }
    
    
    /**
	 * Needed by this abstract class to handle RMAIUtils injection via iPOJO.
	 * In fact, iPOJO currently do not support, abstract class field injection located in another package.
	 */
	protected abstract RMAIUtils getRMAIUtils();
	
	
	/**
	 * Needed by this abstract class to handle Logger injection via iPOJO.
	 * In fact, iPOJO currently do not support, abstract class field injection located in another package.
	 */
	protected abstract Logger getLogger();
	
	
	/**
	 * Get a new instance of RMAIGameLogicManager from extending class.
	 * @param senderASID - RMAI message sender ID
	 * @param recipientASID - RMAI message recipient ID
	 * @return
	 */
	protected abstract Object newRMAIGameLogicManager(short senderASID, short recipientASID);
    
	
	 /*********************************************************************************************/
	/** Listeners : onStartEvent, onEndEvent, onJoinEvent, onQuitEvent, onRMAIEvent, onMooDSEvent**/
	/**********************************************************************************************/

	
	
	public void onStartEvent(StartEvent se) {
		getLogger().log(Logger.INFO,this.getClass(),"OnStartEvent received from ActorSessionID = "+se.getActorSessionID());
	}

	
	/**
	* Event raised when the owner decides to stop the instance.
	**/
	public void onEndEvent(EndEvent ee) {
		getLogger().log(Logger.INFO,this.getClass(),"OnEndEvent received from ActorSessionID = "+ee.getActorSessionID());
	}
	
	
	/**
	* Event raised when a client joined the instance.
	**/
	public void onJoinEvent(JoinEvent je) {
		getLogger().log(Logger.INFO,this.getClass(),"OnJoinEvent received from ActorSessionID = "+je.getActorSessionID());
	}
	
	
	/**
	* Event raised when a client quit the instance.
	**/
	public void onQuitEvent(QuitEvent qe) {
		getLogger().log(Logger.INFO,this.getClass(),"OnQuitEvent received from ActorSessionID = "+qe.getActorSessionID());
	}
	
	
	/**
	 * Event raised when a remote method has been invoked on the game logic.
	 */
	public void onRMAIEvent(RMAIEvent rmaiEvent){
		// declare InvokedMethod and Method needed to execute the RMAI
		Method rmaiMethod;
		InvokedMethod toInvoke;
		// create a RMAIGameLogicManager
		Object rmaiGameLogicManager = newRMAIGameLogicManager(rmaiEvent.getActorSessionID(),rmaiEvent.getRecipientActorSessionID());
		try {
			// deserialize the XML
			toInvoke = getRMAIUtils().readMethodInvocation(rmaiEvent.getInvokedMethodXMLSerialized());
			// get the class of the attribute on which method should be invoked, i.e. RMAIGameLogicManagerImpl class
			Class<?> rmaiGameLogicManagerClass = rmaiGameLogicManager.getClass();
			// get a Method object matching with informations included in toInvoke ...
			rmaiMethod = rmaiGameLogicManagerClass.getMethod(toInvoke.getMethodName(),toInvoke.getParametersTypes());
			// invoke the method on the rmai client manager with the suitable arguments
			rmaiMethod.invoke(rmaiGameLogicManager, toInvoke.getParameters());
		}catch(Throwable t){
			getLogger().log(Logger.WARN,this.getClass(),"Exception on deserialization or on execution of an RMAI",t);
		}
	}
	

	/**
	* Event raised when a client sent MooDS data.
	**/
	public void onMooDSEvent(MooDSEvent me) {
		Hashtable<?,?> table = me.getMooDSData();
		int nbObj = table.size();

		for (int i = 0; i < nbObj; i++) {
			 Object o = (Object) table.get(i + "");
			 treatMooDSMessage(o, me.getActorSessionID());
		 }
	}

	
	/********************************************************************************************************/
	/*************************** Remote Method Asynchronous Invocations Management ***************************/
	/********************************************************************************************************/


	/**
	 * Invoke asynchronously a remote method with no parameters either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @param param4					fifth parameter of the method that should be invoked
	 * @param param5					sixth parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName){
		// create an RMAI Event and send it to application instance (third argument null means that there isn't any parameter in this RMAI)
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, null);
	}


	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameter of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName, Object param0){
		// build the Object array
		Object[] params = {param0};
		// create an RMAI Event and send it to application instance
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, params);
	}


	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName, Object param0,Object param1){
		// build the Object array
		Object[] params = {param0,param1};
		// create an RMAI Event and send it to application instance
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, params);
	}


	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName, Object param0,Object param1, Object param2){
		// build the Object array
		Object[] params = {param0,param1,param2};
		// create an RMAI Event and send it to application instance
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, params);
	}


	/**
	 * Invoke asynchronously a remote method without either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName, Object param0, Object param1, Object param2, Object param3){
		// build the Object array
		Object[] params = {param0,param1,param2,param3};
		// create an RMAI Event and send it to application instance
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, params);
	}


	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @param param4					fifth parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName, Object param0, Object param1, Object param2, Object param3, Object param4){
		// build the Object array
		Object[] params = {param0,param1,param2,param3,param4};
		// create an RMAI Event and send it to application instance
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, params);
	}


	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @param param4					fifth parameter of the method that should be invoked
	 * @param param5					sixth parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName, Object param0, Object param1, Object param2, Object param3, Object param4, Object param5){
		// build the Object array
		Object[] params = {param0,param1,param2,param3,param4,param5};
		// create an RMAI Event and send it to application instance
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, params);
	}


	/**
	 * Invoke asynchronously a remote method either :
	 * <ul>
	 * <li>on a uGASP Client, designated by its actor session id</li>
	 * <li>or on every connected uGASP clients (RMAIClientManager.ActorSessionID.BROADCAST_WITHOUT_GAME_LOGIC.getID())</li>
	 * <li>or on every connected uGASP clients and uGASP game logic (RMAIClientManager.ActorSessionID.BROADCAST_WITH_GAME_LOGIC.getID())</li>
	 * </ul>
	 * Methods that should be invoked must be implemented in the class RMAI client manager.
	 * Only instances methods can be invoked.
	 * Here are the types that can be used as parameters of this remote method:
	 * <ul>
	 * <li>Boolean</li>
	 * <li>Byte</li>
	 * <li>Short</li>
	 * <li>Integer</li>
	 * <li>Long</li>
	 * <li>Double</li>
	 * <li>Float</li>
	 * <li>String</li>
	 * <li>RMAISerializable (you should pay a special attention on recursion : never pass this object reference).</li>
	 * <li>Arrays of those several types</li>
	 * </ul>
	 * 
	 * @param recipientActorSessionID	the actorSession ID of the suitable recipient
	 * @param methodName				the name of the method that should be invoked
	 * @param param0					first parameter of the method that should be invoked
	 * @param param1					second parameter of the method that should be invoked
	 * @param param2					third parameter of the method that should be invoked
	 * @param param3					fourth parameter of the method that should be invoked
	 * @param param4					fifth parameter of the method that should be invoked
	 * @param param5					sixth parameter of the method that should be invoked
	 * @param param6					seventh parameter of the method that should be invoked
	 * @return boolean					true if the response code of the HTTP connection is OK
	 * 									false otherwise
	 */
	public void remoteMethodAsynchronousInvocation(short recipientGASPClientActorSessionID, String methodName, Object param0, Object param1, Object param2, Object param3, Object param4, Object param5, Object param6){
		// build the Object array
		Object[] params = {param0,param1,param2,param3,param4,param5,param6};
		// create an RMAI Event and send it to application instance
		createAndSendRMAIEvent(recipientGASPClientActorSessionID, methodName, params);
	}
	
	/*
	 * This method is used by several remoteMethodAsynchronousInvocation to create an RMAI Event 
	 * with the suitable parameters, and to send it to application instance.
	 */
	private void createAndSendRMAIEvent(short recipientGASPClientActorSessionID, String methodName, Object[] params ){
		String invokedMethodXMLSerialized = null;
		try{
			// Serialization of "parameters" (method name + params, types and values) XMLRPC way
			invokedMethodXMLSerialized = getRMAIUtils().methodInvocation(methodName, params);
		}catch(Throwable t){
			getLogger().log(Logger.WARN,this.getClass(),"Exception on serialization of an RMAI",t);
		}
		
		// create the matching RMAI event
		RMAIEvent rmaiEvent = new RMAIEventImpl(org.ugasp.protocol.rmai.RMAIGameLogicManager.ActorSessionID.uGASP_GAME_LOGIC.getID(),recipientGASPClientActorSessionID,invokedMethodXMLSerialized);
		// raise event
		applicationInstance.onRMAIEvent(rmaiEvent);
	}
	
	
	/************************************************************************/
	/*************************** MooDS Management ***************************/
	/************************************************************************/
	
	
	/**
	 * Send a MooDS Message object (i.e. an object generated by MooDS) from a client to another one.
	 * 
	 * @param o the object to send
	 * @param senderActorSessionID the actorSessionID of the sender
	 * @param destinationActorSessionID the actorSessionID of the recipient
	 */
	public void sendMooDSMessageToPlayer(Object moodsObject, short senderActorSessionID, short destinationActorSessionID) {
				
		Hashtable<String,Object> table = new Hashtable<String,Object>();
		// insert object at the key "0" --> only one element is present in the map
		table.put("0", moodsObject);
		// create the corresponding dataEvent
		MooDSEvent moodsEvent = new MooDSEventImpl(senderActorSessionID, table);
		// send it to the recipient client designated by its actorSessionID
		applicationInstance.sendMooDSDataToActorSession(destinationActorSessionID,moodsEvent);
	
	}
	
	
	/**
	 * Treatment of MooDS object received.
	 * You have to insert you treatment for each object after checking
	 * from which class Object o is an instance.
	 *
 	 *@param senderActorSessionID the actorSessionID of the sender of datas
	 *@param Object	the object to treat. 
	 */
	protected void treatMooDSMessage(Object moodsObject, short senderActorSessionID){
		System.err.println("Treatment of MooDS messages is not defined by game logic: " +
				"treatMooDSMessage(Object moodsObject, short senderActorSessionID) method " +
				"should be overrided in your subclass of GameLogicImpl if you want to use MooDS protocol.");
	}
	
	 /**
     * Get the CustomTypes attribute of this game logic.
     */
    public CustomTypes getCustomTypes() {
		return customTypes;
	}
    
    
    /**
	 * Set the CustomTypes for this game logic.
	 * Each game logic could have one subclass of CustomTypes.
	 * If you want to use MooDS protocol, this method HAVE to be called 
	 * in the constructor of your customized game logic.
	 * 
	 * @param an instance of CustomTypes implementation.
	 */
    public void setCustomTypes(CustomTypes customTypes){
    	this.customTypes = customTypes;
    }
    
}